from __future__ import annotations

import argparse
import re
import time
from contextlib import nullcontext
from operator import eq, gt, lt
from typing import TYPE_CHECKING, Any
from unittest.mock import Mock, call, patch

import freezegun
import pytest
import requests.cookies

from streamlink.logger import StreamlinkLogger
from streamlink.options import Argument, Options
from streamlink.plugin import (
    HIGH_PRIORITY,
    NORMAL_PRIORITY,
    Plugin,
    PluginArgument,
    PluginArguments,
    PluginError,
    pluginargument,
    pluginmatcher,
)

# noinspection PyProtectedMember
from streamlink.plugin.plugin import (
    _COOKIE_KEYS,  # noqa: PLC2701
    _PLUGINARGUMENT_TYPE_REGISTRY,  # noqa: PLC2701
    Matcher,
    parse_params,
    stream_weight,
)


if TYPE_CHECKING:
    from streamlink.session import Streamlink


class FakePlugin(Plugin):
    def _get_streams(self):
        pass  # pragma: no cover


class RenamedPlugin(FakePlugin):
    __module__ = "foo.bar.baz"


class CustomConstructorOnePlugin(FakePlugin):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class CustomConstructorTwoPlugin(FakePlugin):
    def __init__(self, session, url):
        super().__init__(session, url)


class TestPlugin:
    @pytest.mark.parametrize(
        ("pluginclass", "module", "logger"),
        [
            (FakePlugin, "test_plugin", "streamlink.tests.test_plugin"),
            (RenamedPlugin, "baz", "streamlink.foo.bar.baz"),
            (CustomConstructorOnePlugin, "test_plugin", "streamlink.tests.test_plugin"),
            (CustomConstructorTwoPlugin, "test_plugin", "streamlink.tests.test_plugin"),
        ],
    )
    def test_constructor(
        self,
        monkeypatch: pytest.MonkeyPatch,
        caplog: pytest.LogCaptureFixture,
        pluginclass: type[Plugin],
        module: str,
        logger: str,
        session: Streamlink,
    ):
        mock_cache = Mock()
        mock_load_cookies = Mock()
        monkeypatch.setattr("streamlink.plugin.plugin.Cache", mock_cache)
        monkeypatch.setattr(pluginclass, "load_cookies", mock_load_cookies)

        plugin = pluginclass(session, "http://localhost")

        assert not caplog.records

        assert plugin.session is session
        assert plugin.url == "http://localhost"

        assert plugin.module == module

        assert isinstance(plugin.logger, StreamlinkLogger)
        assert plugin.logger.name == logger

        assert mock_cache.call_args_list == [call(filename="plugin-cache.json", key_prefix=module, disabled=False)]
        assert plugin.cache == mock_cache()

        assert mock_load_cookies.call_args_list == [call()]

    def test_constructor_options(self):
        one = FakePlugin(Mock(), "https://mocked", Options({"key": "val"}))
        two = FakePlugin(Mock(), "https://mocked")
        assert one.get_option("key") == "val"
        assert two.get_option("key") is None

        one.set_option("key", "other")
        assert one.get_option("key") == "other"
        assert two.get_option("key") is None

    @pytest.mark.parametrize(
        ("session", "expected"),
        [
            pytest.param({}, False, id="default"),
            pytest.param({"no-plugin-cache": False}, False, id="not-disabled"),
            pytest.param({"no-plugin-cache": True}, True, id="disabled"),
        ],
        indirect=["session"],
    )
    def test_disabled_cache(self, session: Streamlink, expected: bool):
        plugin = FakePlugin(session, "https://mocked")
        assert plugin.cache._disabled is expected


class TestPluginMatcher:
    # noinspection PyUnusedLocal
    def test_decorator(self):
        class MyPlugin:
            pass

        with pytest.raises(TypeError) as cm:
            pluginmatcher(re.compile(r""))(MyPlugin)  # type: ignore

        assert str(cm.value) == "MyPlugin is not a Plugin"

    def test_named_duplicate(self):
        class MyPlugin(FakePlugin):
            pass

        matcher = pluginmatcher(re.compile(r"http://foo"), name="foo")

        with pytest.raises(ValueError, match=r"^A matcher named 'foo' has already been registered$"):
            matcher(matcher(MyPlugin))

    def test_no_matchers(self):
        class MyPlugin(FakePlugin):
            pass

        plugin = MyPlugin(Mock(), "http://foo")
        assert plugin.url == "http://foo"
        assert plugin.matchers == []
        assert plugin.matches == []
        assert plugin.matcher is None
        assert plugin.match is None

    def test_matchers(self):
        @pluginmatcher(re.compile(r"foo", re.VERBOSE))
        @pluginmatcher(re.compile(r"bar"), priority=HIGH_PRIORITY)
        @pluginmatcher(re.compile(r"baz"), priority=HIGH_PRIORITY, name="baz")
        class MyPlugin(FakePlugin):
            pass

        assert MyPlugin.matchers == [
            Matcher(re.compile(r"foo", re.VERBOSE), NORMAL_PRIORITY),
            Matcher(re.compile(r"bar"), HIGH_PRIORITY),
            Matcher(re.compile(r"baz"), HIGH_PRIORITY, "baz"),
        ]

    # noinspection PyAbstractClass
    def test_matchers_inheritance(self):
        @pluginmatcher(re.compile(r"foo"))
        @pluginmatcher(re.compile(r"bar"))
        class PluginOne(FakePlugin):
            pass

        @pluginmatcher(re.compile(r"baz"))
        @pluginmatcher(re.compile(r"qux"))
        class PluginTwo(PluginOne):
            pass

        assert PluginOne.matchers is not PluginTwo.matchers
        assert PluginOne.matchers == [
            Matcher(re.compile(r"foo"), NORMAL_PRIORITY),
            Matcher(re.compile(r"bar"), NORMAL_PRIORITY),
        ]
        assert PluginTwo.matchers == [
            Matcher(re.compile(r"baz"), NORMAL_PRIORITY),
            Matcher(re.compile(r"qux"), NORMAL_PRIORITY),
            Matcher(re.compile(r"foo"), NORMAL_PRIORITY),
            Matcher(re.compile(r"bar"), NORMAL_PRIORITY),
        ]

    # noinspection PyUnusedLocal,PyAbstractClass
    def test_matchers_inheritance_named_duplicate(self):
        @pluginmatcher(name="foo", pattern=re.compile(r"foo"))
        class PluginOne(FakePlugin):
            pass

        with pytest.raises(ValueError, match=r"^A matcher named 'foo' has already been registered$"):

            @pluginmatcher(name="foo", pattern=re.compile(r"foo"))
            class PluginTwo(PluginOne):
                pass

    def test_matchers_not_matching(self):
        @pluginmatcher(re.compile(r"http://foo"))
        class MyPlugin(FakePlugin):
            pass

        with pytest.raises(PluginError, match=r"^The input URL did not match any of this plugin's matchers$"):
            MyPlugin(Mock(), "http://bar")

    def test_no_matchers_not_matching(self):
        plugin = FakePlugin(Mock(), "")
        assert plugin.match is None

    def test_url_setter(self):
        @pluginmatcher(re.compile(r"http://(foo)"))
        @pluginmatcher(re.compile(r"http://(bar)"))
        @pluginmatcher(re.compile(r"http://(baz)"))
        class MyPlugin(FakePlugin):
            pass

        plugin = MyPlugin(Mock(), "http://foo")
        assert plugin.url == "http://foo"
        assert [m is not None for m in plugin.matches] == [True, False, False]
        assert plugin.matcher is plugin.matchers[0].pattern
        assert plugin.match is not None
        assert plugin.match.group(1) == "foo"

        plugin.url = "http://bar"
        assert plugin.url == "http://bar"
        assert [m is not None for m in plugin.matches] == [False, True, False]
        assert plugin.matcher is plugin.matchers[1].pattern
        assert plugin.match is not None
        assert plugin.match.group(1) == "bar"

        plugin.url = "http://baz"
        assert plugin.url == "http://baz"
        assert [m is not None for m in plugin.matches] == [False, False, True]
        assert plugin.matcher is plugin.matchers[2].pattern
        assert plugin.match is not None
        assert plugin.match.group(1) == "baz"

        with pytest.raises(PluginError, match=r"^The input URL did not match any of this plugin's matchers$"):
            plugin.url = "http://qux"
        assert plugin.url == "http://qux"
        assert [m is not None for m in plugin.matches] == [False, False, False]
        assert plugin.matcher is None
        assert plugin.match is None

    def test_named_matchers_and_matches(self):
        @pluginmatcher(re.compile(r"http://foo"), name="foo")
        @pluginmatcher(re.compile(r"http://bar"), name="bar")
        class MyPlugin(FakePlugin):
            pass

        plugin = MyPlugin(Mock(), "http://foo")

        assert plugin.matchers["foo"] is plugin.matchers[0]
        assert plugin.matchers["bar"] is plugin.matchers[1]
        with pytest.raises(IndexError):
            plugin.matchers.__getitem__(2)
        with pytest.raises(KeyError):
            plugin.matchers.__getitem__("baz")

        assert plugin.matches["foo"] is plugin.matches[0]
        assert plugin.matches["bar"] is plugin.matches[1]
        assert plugin.matches["foo"] is not None
        assert plugin.matches["bar"] is None
        with pytest.raises(IndexError):
            plugin.matches.__getitem__(2)
        with pytest.raises(KeyError):
            plugin.matches.__getitem__("baz")

        plugin.url = "http://bar"
        assert plugin.matches["foo"] is None
        assert plugin.matches["bar"] is not None

        with pytest.raises(PluginError, match=r"^The input URL did not match any of this plugin's matchers$"):
            plugin.url = "http://baz"
        assert plugin.matches["foo"] is None
        assert plugin.matches["bar"] is None


class TestPluginArguments:
    @pluginargument("foo", dest="_foo", help="FOO")
    @pluginargument("bar", dest="_bar", help="BAR")
    @pluginargument("baz", dest="_baz", help="BAZ")
    class DecoratedPlugin(FakePlugin):
        pass

    class ClassAttrPlugin(FakePlugin):
        arguments = PluginArguments(
            PluginArgument("foo", dest="_foo", help="FOO"),
            PluginArgument("bar", dest="_bar", help="BAR"),
            PluginArgument("baz", dest="_baz", help="BAZ"),
        )

    def test_pluginargument_type_registry(self):
        assert _PLUGINARGUMENT_TYPE_REGISTRY
        assert all(callable(value) for value in _PLUGINARGUMENT_TYPE_REGISTRY.values())

    @pytest.mark.parametrize("pluginclass", [DecoratedPlugin, ClassAttrPlugin])
    def test_arguments(self, pluginclass: type[Plugin]):
        assert pluginclass.arguments is not None
        assert tuple(arg.name for arg in pluginclass.arguments) == ("foo", "bar", "baz"), "Argument name"
        assert tuple(arg.dest for arg in pluginclass.arguments) == ("_foo", "_bar", "_baz"), "Argument keyword"
        assert tuple(arg.options.get("help") for arg in pluginclass.arguments) == ("FOO", "BAR", "BAZ"), "argparse keyword"
        assert isinstance(pluginclass.get_argument("foo"), Argument)
        assert pluginclass.get_argument("doesnotexist") is None
        assert Plugin.get_argument("no-arguments") is None

    @pytest.mark.parametrize("pluginclass", [DecoratedPlugin, ClassAttrPlugin])
    def test_arguments_mixed(self, pluginclass):
        @pluginargument("qux")
        class MixedPlugin(pluginclass):
            pass

        assert tuple(arg.name for arg in MixedPlugin.arguments) == ("qux", "foo", "bar", "baz")

    def test_arguments_inheritance(self):
        @pluginargument("foo", help="FOO")
        @pluginargument("bar", help="BAR")
        class PluginOne(FakePlugin):
            pass

        @pluginargument("baz", help="BAZ")
        @pluginargument("qux", help="QUX")
        class PluginTwo(PluginOne):
            pass

        assert PluginOne.arguments is not PluginTwo.arguments
        assert tuple(arg.name for arg in PluginOne.arguments) == ("foo", "bar")
        assert tuple(arg.name for arg in PluginTwo.arguments) == ("baz", "qux", "foo", "bar")

    @pytest.mark.parametrize(
        ("options", "args", "expected", "raises"),
        [
            pytest.param(
                {"type": "int"},
                ["--myplugin-foo", "123"],
                123,
                nullcontext(),
                id="int",
            ),
            pytest.param(
                {"type": "float"},
                ["--myplugin-foo", "123.456"],
                123.456,
                nullcontext(),
                id="float",
            ),
            pytest.param(
                {"type": "bool"},
                ["--myplugin-foo", "yes"],
                True,
                nullcontext(),
                id="bool",
            ),
            pytest.param(
                {"type": "keyvalue"},
                ["--myplugin-foo", "key=value"],
                ("key", "value"),
                nullcontext(),
                id="keyvalue",
            ),
            pytest.param(
                {"type": "comma_list_filter", "type_args": (["one", "two", "four"],)},
                ["--myplugin-foo", "one,two,three,four"],
                ["one", "two", "four"],
                nullcontext(),
                id="comma_list_filter - args",
            ),
            pytest.param(
                {"type": "comma_list_filter", "type_kwargs": {"acceptable": ["one", "two", "four"]}},
                ["--myplugin-foo", "one,two,three,four"],
                ["one", "two", "four"],
                nullcontext(),
                id="comma_list_filter - kwargs",
            ),
            pytest.param(
                {"type": "hours_minutes_seconds"},
                ["--myplugin-foo", "1h2m3s"],
                3723,
                nullcontext(),
                id="hours_minutes_seconds",
            ),
            pytest.param(
                {"type": "UNKNOWN"},
                None,
                None,
                pytest.raises(TypeError),
                id="UNKNOWN",
            ),
        ],
    )
    def test_type_argument_map(self, options: dict, args: list, expected: Any, raises: nullcontext):
        class MyPlugin(FakePlugin):
            pass

        with raises:
            pluginargument("foo", **options)(MyPlugin)
            assert MyPlugin.arguments is not None
            pluginarg = MyPlugin.arguments.get("foo")
            assert pluginarg

            parser = argparse.ArgumentParser()
            parser.add_argument(pluginarg.argument_name("myplugin"), **pluginarg.options)
            namespace = parser.parse_args(args)
            assert namespace.myplugin_foo == expected

    def test_decorator_typeerror(self, monkeypatch: pytest.MonkeyPatch):
        class NotAPluginMeta(type):
            def __repr__(self) -> str:
                return self.__name__

        with pytest.raises(TypeError, match=r"^NotAPlugin is not a Plugin$"):
            # noinspection PyUnusedLocal
            @pluginargument("foo")  # type: ignore[arg-type, ty:invalid-argument-type]
            class NotAPlugin(metaclass=NotAPluginMeta):
                pass

    def test_empty(self):
        assert FakePlugin.arguments is not None
        assert tuple(iter(FakePlugin.arguments)) == ()


@pytest.mark.parametrize("attr", ["id", "author", "category", "title"])
def test_plugin_metadata(attr):
    plugin = FakePlugin(Mock(), "https://foo.bar/")
    getter = getattr(plugin, f"get_{attr}")
    assert callable(getter)

    assert getattr(plugin, attr) is None
    assert getter() is None

    setattr(plugin, attr, " foo bar ")
    assert getter() == "foo bar"

    class Foo:
        def __str__(self):
            return " baz qux "

    setattr(plugin, attr, Foo())
    assert getter() == "baz qux"


class TestCookies:
    @staticmethod
    def create_cookie_dict(name, value, expires=None):
        return dict(
            version=0,
            name=name,
            value=value,
            port=None,
            domain="test.se",
            path="/",
            secure=False,
            expires=expires,
            discard=True,
            comment=None,
            comment_url=None,
            rest={"HttpOnly": None},
            rfc2109=False,
        )

    @pytest.fixture()
    def pluginclass(self):
        class MyPlugin(FakePlugin):
            __module__ = "myplugin"

        return MyPlugin

    @pytest.fixture()
    def plugincache(self, request):
        with patch("streamlink.plugin.plugin.Cache") as mock_cache:
            cache = mock_cache("plugin-cache.json", "myplugin")
            cache.get_all.return_value = request.param
            yield cache

    @pytest.fixture()
    def logger(self, monkeypatch: pytest.MonkeyPatch, pluginclass: type[Plugin]):
        mock_getlogger = Mock()
        monkeypatch.setattr("streamlink.plugin.plugin.getLogger", mock_getlogger)
        return mock_getlogger(pluginclass.__module__)

    @pytest.fixture()
    def plugin(self, pluginclass: type[Plugin], session: Streamlink, plugincache: Mock, logger: Mock):
        plugin = pluginclass(session, "http://test.se")
        assert plugin.cache is plugincache
        assert plugin.logger is logger
        return plugin

    @staticmethod
    def _cookie_to_dict(cookie):
        r = {name: getattr(cookie, name, None) for name in _COOKIE_KEYS}
        r["rest"] = getattr(cookie, "rest", getattr(cookie, "_rest", None))
        return r

    def _cookies_to_list(self, cookies):
        return [self._cookie_to_dict(cookie) for cookie in cookies]

    @pytest.mark.parametrize(
        "plugincache",
        [
            {
                "__cookie:test-name1:test.se:80:/": create_cookie_dict("test-name1", "test-value1"),
                "__cookie:test-name2:test.se:80:/": create_cookie_dict("test-name2", "test-value2"),
                "unrelated": "data",
            },
        ],
        indirect=True,
    )
    def test_load(self, session: Streamlink, plugin: Plugin, plugincache: Mock, logger: Mock):
        assert self._cookies_to_list(session.http.cookies) == self._cookies_to_list([
            requests.cookies.create_cookie("test-name1", "test-value1", domain="test.se"),
            requests.cookies.create_cookie("test-name2", "test-value2", domain="test.se"),
        ])
        assert logger.debug.call_args_list == [call("Restored cookies: test-name1, test-name2")]

    @pytest.mark.parametrize("plugincache", [{}], indirect=True)
    def test_save(self, session: Streamlink, plugin: Plugin, plugincache: Mock, logger: Mock):
        cookie1 = requests.cookies.create_cookie("test-name1", "test-value1", domain="test.se")
        cookie2 = requests.cookies.create_cookie("test-name2", "test-value2", domain="test.se")
        session.http.cookies.set_cookie(cookie1)
        session.http.cookies.set_cookie(cookie2)

        plugin.save_cookies(lambda cookie: cookie.name == "test-name1", default_expires=3600)
        assert plugincache.set.call_args_list == [
            call(
                "__cookie:test-name1:test.se:80:/",
                self.create_cookie_dict("test-name1", "test-value1", None),
                3600,
            ),
        ]
        assert logger.debug.call_args_list == [call("Saved cookies: test-name1")]

    @freezegun.freeze_time("1970-01-01T00:00:00Z")
    @pytest.mark.parametrize("plugincache", [{}], indirect=True)
    def test_save_expires(self, session: Streamlink, plugin: Plugin, plugincache: Mock):
        cookie = requests.cookies.create_cookie(
            "test-name",
            "test-value",
            domain="test.se",
            expires=time.time() + 3600,
            rest={"HttpOnly": None},
        )
        session.http.cookies.set_cookie(cookie)

        plugin.save_cookies(default_expires=60)
        assert plugincache.set.call_args_list == [
            call(
                "__cookie:test-name:test.se:80:/",
                self.create_cookie_dict("test-name", "test-value", 3600),
                3600,
            ),
        ]

    @pytest.mark.parametrize(
        "plugincache",
        [
            {
                "__cookie:test-name1:test.se:80:/": create_cookie_dict("test-name1", "test-value1", None),
                "__cookie:test-name2:test.se:80:/": create_cookie_dict("test-name2", "test-value2", None),
                "unrelated": "data",
            },
        ],
        indirect=True,
    )
    def test_clear(self, session: Streamlink, plugin: Plugin, plugincache: Mock):
        assert tuple(session.http.cookies.keys()) == ("test-name1", "test-name2")

        plugin.clear_cookies()
        assert call("__cookie:test-name1:test.se:80:/", None, 0) in plugincache.set.call_args_list
        assert call("__cookie:test-name2:test.se:80:/", None, 0) in plugincache.set.call_args_list
        assert len(session.http.cookies.keys()) == 0

    @pytest.mark.parametrize(
        "plugincache",
        [
            {
                "__cookie:test-name1:test.se:80:/": create_cookie_dict("test-name1", "test-value1", None),
                "__cookie:test-name2:test.se:80:/": create_cookie_dict("test-name2", "test-value2", None),
                "unrelated": "data",
            },
        ],
        indirect=True,
    )
    def test_clear_filter(self, session: Streamlink, plugin: Plugin, plugincache: Mock):
        assert tuple(session.http.cookies.keys()) == ("test-name1", "test-name2")

        plugin.clear_cookies(lambda cookie: cookie.name == "test-name2")
        assert call("__cookie:test-name1:test.se:80:/", None, 0) not in plugincache.set.call_args_list
        assert call("__cookie:test-name2:test.se:80:/", None, 0) in plugincache.set.call_args_list
        assert tuple(session.http.cookies.keys()) == ("test-name1",)


@pytest.mark.parametrize(
    ("params", "expected"),
    [
        (
            None,
            {},
        ),
        (
            "foo=bar",
            dict(foo="bar"),
        ),
        (
            "verify=False",
            dict(verify=False),
        ),
        (
            "timeout=123.45",
            dict(timeout=123.45),
        ),
        (
            "verify=False params={'key': 'a value'}",
            dict(verify=False, params=dict(key="a value")),
        ),
        (
            "\"conn=['B:1', 'S:authMe', 'O:1', 'NN:code:1.23', 'NS:flag:ok', 'O:0']",
            dict(conn=["B:1", "S:authMe", "O:1", "NN:code:1.23", "NS:flag:ok", "O:0"]),
        ),
    ],
)
def test_parse_params(params, expected):
    assert parse_params(params) == expected


@pytest.mark.parametrize(
    ("weight", "expected"),
    [
        ("720p", (720, "pixels")),
        ("720p+", (721, "pixels")),
        ("720p60", (780, "pixels")),
    ],
)
def test_stream_weight_value(weight, expected):
    assert stream_weight(weight) == expected


@pytest.mark.parametrize(
    ("weight_a", "operator", "weight_b"),
    [
        ("720p+", gt, "720p"),
        ("720p_3000k", gt, "720p_2500k"),
        ("720p60_3000k", gt, "720p_3000k"),
        ("3000k", gt, "2500k"),
        ("720p", eq, "720p"),
        ("720p_3000k", lt, "720p+_3000k"),
        # with audio
        ("720p+a256k", gt, "720p+a128k"),
        ("720p+a256k", gt, "360p+a256k"),
        ("720p+a128k", gt, "360p+a256k"),
    ],
)
def test_stream_weight(weight_a, weight_b, operator):
    assert operator(stream_weight(weight_a), stream_weight(weight_b))



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING
from unittest.mock import call, patch

import pytest

from streamlink_cli.utils.player import find_default_player


if TYPE_CHECKING:
    from unittest.mock import Mock


# noinspection PyTestParametrized
class TestFindDefaultPlayer:
    @pytest.fixture(autouse=True)
    def _fakehome(self, request: pytest.FixtureRequest, monkeypatch: pytest.MonkeyPatch):
        home = getattr(request, "param", "")
        monkeypatch.setattr("streamlink_cli.utils.player.Path.home", lambda: home)

    @pytest.fixture(autouse=True)
    def _environ(self, request: pytest.FixtureRequest, monkeypatch: pytest.MonkeyPatch):
        for key, value in getattr(request, "param", {}).items():
            monkeypatch.setenv(key, value)

    @pytest.fixture(autouse=True)
    def which(self, request: pytest.FixtureRequest):
        return_values = getattr(request, "param", {})
        with patch("streamlink_cli.utils.player.which", side_effect=lambda path: return_values.get(path, None)) as mock_which:
            yield mock_which

    @pytest.fixture(autouse=True)
    def _assert_find_default_player(
        self,
        request: pytest.FixtureRequest,
        monkeypatch: pytest.MonkeyPatch,
        which: Mock,
        lookups: list[str],
        expected: str | None,
    ):
        monkeypatch.setattr("streamlink_cli.utils.player.is_win32", "win32" in request.function.__name__)
        monkeypatch.setattr("streamlink_cli.utils.player.is_darwin", "darwin" in request.function.__name__)
        yield
        assert find_default_player() == expected
        assert which.call_args_list == [call(path) for path in lookups]

    @pytest.mark.windows_only()
    @pytest.mark.parametrize(
        "_environ",
        [
            {
                "PROGRAMFILES": "C:\\Program Files",
                "PROGRAMFILES(X86)": "C:\\Program Files (x86)",
                "PROGRAMW6432": "C:\\Program Files",
            },
        ],
        indirect=True,
    )
    @pytest.mark.parametrize(
        ("which", "lookups", "expected"),
        [
            pytest.param(
                {"vlc.exe": "C:\\Program Files\\VideoLAN\\VLC\\vlc.exe"},
                ["vlc.exe"],
                Path("C:\\Program Files\\VideoLAN\\VLC\\vlc.exe"),
                id="PATH lookup success",
            ),
            pytest.param(
                {"C:\\Program Files (x86)\\VideoLAN\\VLC\\vlc.exe": "C:\\Program Files (x86)\\VideoLAN\\VLC\\vlc.exe"},
                [
                    "vlc.exe",
                    "C:\\Program Files\\VideoLAN\\VLC\\vlc.exe",
                    "C:\\Program Files (x86)\\VideoLAN\\VLC\\vlc.exe",
                ],
                Path("C:\\Program Files (x86)\\VideoLAN\\VLC\\vlc.exe"),
                id="fallback paths lookup success",
            ),
            pytest.param(
                {},
                [
                    "vlc.exe",
                    "C:\\Program Files\\VideoLAN\\VLC\\vlc.exe",
                    "C:\\Program Files (x86)\\VideoLAN\\VLC\\vlc.exe",
                ],
                None,
                id="lookup failure",
            ),
        ],
        indirect=["which"],
    )
    def test_win32(self):
        pass

    @pytest.mark.windows_only()
    @pytest.mark.parametrize(
        "_environ",
        [
            {
                "PROGRAMFILES": "",
                "PROGRAMFILES(X86)": "",
                "PROGRAMW6432": "",
            },
        ],
        indirect=True,
    )
    @pytest.mark.parametrize(
        ("which", "lookups", "expected"),
        [
            pytest.param(
                {"vlc.exe": "C:\\Program Files\\VideoLAN\\VLC\\vlc.exe"},
                ["vlc.exe"],
                Path("C:\\Program Files\\VideoLAN\\VLC\\vlc.exe"),
                id="PATH lookup success",
            ),
            pytest.param(
                {},
                ["vlc.exe"],
                None,
                id="no fallback paths",
            ),
        ],
        indirect=["which"],
    )
    def test_win32_no_env_vars(self):
        pass

    @pytest.mark.posix_only()
    @pytest.mark.usefixtures("_fakehome")
    @pytest.mark.parametrize("_fakehome", ["/Users/fake"], indirect=True)
    @pytest.mark.parametrize(
        ("which", "lookups", "expected"),
        [
            pytest.param(
                {"vlc": "/usr/bin/vlc"},
                ["VLC", "vlc"],
                Path("/usr/bin/vlc"),
                id="PATH lookup success",
            ),
            pytest.param(
                {"/Applications/VLC.app/Contents/MacOS/VLC": "/Applications/VLC.app/Contents/MacOS/VLC"},
                [
                    "VLC",
                    "vlc",
                    "/Applications/VLC.app/Contents/MacOS/VLC",
                ],
                Path("/Applications/VLC.app/Contents/MacOS/VLC"),
                id="fallback paths lookup success",
            ),
            pytest.param(
                {},
                [
                    "VLC",
                    "vlc",
                    "/Applications/VLC.app/Contents/MacOS/VLC",
                    "/Applications/VLC.app/Contents/MacOS/vlc",
                    "/Users/fake/Applications/VLC.app/Contents/MacOS/VLC",
                    "/Users/fake/Applications/VLC.app/Contents/MacOS/vlc",
                ],
                None,
                id="lookup failure",
            ),
        ],
        indirect=["which"],
    )
    def test_darwin(self):
        pass

    @pytest.mark.posix_only()
    @pytest.mark.parametrize(
        ("which", "lookups", "expected"),
        [
            pytest.param(
                {"vlc": "/usr/bin/vlc"},
                ["vlc"],
                Path("/usr/bin/vlc"),
                id="lookup success",
            ),
            pytest.param(
                {},
                ["vlc"],
                None,
                id="lookup failure",
            ),
        ],
        indirect=["which"],
    )
    def test_other(self):
        pass



from __future__ import annotations

import abc
import ast
import pkgutil
import re
import tokenize
from pathlib import Path
from typing import TYPE_CHECKING, Any, ClassVar

from docutils import nodes
from docutils.parsers.rst import Directive
from docutils.statemachine import StringList
from sphinx.errors import ExtensionError
from sphinx.util.nodes import nested_parse_with_titles

from streamlink import plugins as streamlink_plugins


if TYPE_CHECKING:
    from collections.abc import Iterator


class IDatalistItem(abc.ABC):
    @abc.abstractmethod
    def generate(self) -> Iterator[str]:
        raise NotImplementedError


class IMetadataItem(IDatalistItem):
    @abc.abstractmethod
    def set(self, value: Any) -> None:
        raise NotImplementedError


class MetadataItem(IMetadataItem):
    def __init__(self, title: str):
        self.title = title
        self.value: str | None = None

    def set(self, value: str) -> None:
        self.value = value

    def generate(self) -> Iterator[str]:
        if self.value is None:
            return
        yield f":{self.title}: {self.value}"


class MetadataList(IMetadataItem):
    def __init__(self, title: str):
        self.title = title
        self.value: list[str] = []

    def set(self, value: str) -> None:
        self.value.append(value)

    def generate(self) -> Iterator[str]:
        if not self.value:
            return
        yield f":{self.title}: - {' '.join(self.get_item(0))}"
        indent = " " * len(f":{self.title}:")
        for idx in range(1, len(self.value)):
            yield f"{indent} - {' '.join(self.get_item(idx))}"

    def get_item(self, idx: int) -> Iterator[str]:
        yield self.value[idx]


class MetadataWebbrowserItem(MetadataItem):
    def __init__(self):
        super().__init__("Web browser")

    def generate(self) -> Iterator[str]:
        if self.value is None:
            return
        yield from super().generate()
        yield " [:ref:`? <cli:Web browser options>`]"


class MetadataMetadataList(MetadataList):
    def __init__(self):
        super().__init__("Metadata")

    def get_item(self, idx: int) -> Iterator[str]:
        variable, *data = str(self.value[idx]).split(" ")
        yield f":ref:`{variable} <cli/metadata:Variables>`"
        if data:
            yield " ".join(["-", *data])


class PluginOnGithub(IDatalistItem):
    url_source = "https://github.com/streamlink/streamlink/blob/master/src/streamlink/plugins/{name}.py"
    url_issues = "https://github.com/streamlink/streamlink/issues?q=is%3Aissue+sort%3Aupdated-desc+plugins.{name}"

    def __init__(self, pluginname: str):
        self.pluginname = pluginname

    def generate(self) -> Iterator[str]:
        source = f"`Source <{self.url_source.format(name=self.pluginname)}>`__"
        issues = f"`Issues <{self.url_issues.format(name=self.pluginname)}>`__"
        yield f":GitHub: {source}, {issues}"


class PluginArguments(ast.NodeVisitor, IDatalistItem):
    def __init__(self, pluginname, pluginast):
        super().__init__()
        self.pluginname = pluginname
        self.arguments = []
        self.visit(pluginast)

    def generate(self) -> Iterator[str]:
        if not self.arguments:
            return
        indent = " " * len(":Arguments: ")
        yield f":Arguments: - :option:`--{self.arguments[0]}`"
        for arg in self.arguments[1:]:
            yield f"{indent} - :option:`--{arg}`"

    def visit_ClassDef(self, node: ast.ClassDef) -> None:
        for base in node.bases:
            if not isinstance(base, ast.Name):
                continue
            if base.id == "Plugin":
                break
        else:
            return

        for decorator in node.decorator_list:
            if (
                not isinstance(decorator, ast.Call)
                or not isinstance(decorator.func, ast.Name)
                or decorator.func.id != "pluginargument"
                or (len(decorator.args) == 0 and len(decorator.keywords) == 0)
            ):
                continue

            if any(
                True
                for kw in decorator.keywords
                if (
                    kw.arg == "help"
                    and type(kw.value) is ast.Attribute
                    and kw.value.attr == "SUPPRESS"
                    and type(kw.value.value) is ast.Name
                    and kw.value.value.id == "argparse"
                )
            ):
                continue

            custom_name = next(
                (
                    str(kw.value.value)
                    for kw in decorator.keywords
                    if kw.arg == "argument_name" and type(kw.value) is ast.Constant
                ),
                None,
            )  # fmt: skip
            if custom_name:
                self.arguments.append(custom_name)
                continue

            name = next(
                (
                    str(kw.value.value)
                    for kw in decorator.keywords
                    if kw.arg == "name" and type(kw.value) is ast.Constant
                ),
                None,
            )  # fmt: skip
            if not name and decorator.args and type(decorator.args[0]) is ast.Constant:
                name = str(decorator.args[0].value)
            if name:
                self.arguments.append(f"{self.pluginname}-{name}")


class PluginMetadata:
    def __init__(self, name: str, pluginast):
        self.name: str = name
        self.items: dict[str, IMetadataItem] = dict(
            description=MetadataItem("Description"),
            url=MetadataList("URL(s)"),
            type=MetadataItem("Type"),
            webbrowser=MetadataWebbrowserItem(),
            metadata=MetadataMetadataList(),
            region=MetadataItem("Region"),
            account=MetadataItem("Account"),
            notes=MetadataList("Notes"),
        )
        self.additional: list[IDatalistItem] = [
            PluginArguments(name, pluginast),
            PluginOnGithub(name),
        ]

    def set(self, key: str, value: str) -> None:
        if key not in self.items:
            raise KeyError(f"Invalid plugin metadata key '{key}' in plugin '{self.name}'")
        self.items[key].set(value)

    def generate(self) -> Iterator[str]:
        yield self.name
        yield "-" * len(self.name)
        yield ""
        for metadata in self.items.values():
            yield from metadata.generate()
        for item in self.additional:
            yield from item.generate()
        yield ""
        yield ""


class PluginFinder:
    _re_metadata_item = re.compile(r"\n\$(\w+) (.+)(?=\n\$|$)", re.MULTILINE)

    protocol_plugins: ClassVar[list[str]] = [
        "http",
        "hls",
        "dash",
    ]

    def __init__(self):
        plugins_path = Path(streamlink_plugins.__path__[0])
        self.plugins = [
            (pname, plugins_path / f"{pname}.py")
            for finder, pname, ispkg in pkgutil.iter_modules([str(plugins_path)])
            if not pname.startswith("common_") and pname not in self.protocol_plugins
        ]

    def get_plugins(self):
        for pluginname, pluginfile in self.plugins:
            pluginmetadata = self._parse_plugin(pluginname, pluginfile)
            if pluginmetadata:
                yield pluginmetadata

    def _parse_plugin(self, pluginname: str, pluginfile: Path) -> PluginMetadata | None:
        with pluginfile.open("r", encoding="utf-8") as handle:
            # read until the first token has been parsed
            for tokeninfo in tokenize.generate_tokens(handle.readline):
                # the very first token needs to be a string / block comment with the metadata
                if tokeninfo.type != tokenize.STRING or not self._re_metadata_item.search(tokeninfo.string):
                    return None
                metadata = tokeninfo.string.strip()
                break

            try:
                # continue reading the plugin file with the same handle
                pluginsource = handle.read()
                # build AST from plugin source for finding the used plugin arguments
                pluginast = ast.parse(pluginsource, str(pluginfile))

                pluginmetadata = PluginMetadata(pluginname, pluginast)
                for item in self._re_metadata_item.finditer(metadata):
                    key, value = item.groups()
                    pluginmetadata.set(key, value)

                return pluginmetadata

            except Exception as err:
                raise ExtensionError(f"Error while parsing plugin file {pluginfile.name}", err) from err


class PluginsDirective(Directive):
    # noinspection PyMethodMayBeStatic
    def generate(self):
        pluginfinder = PluginFinder()

        for pluginmetadata in pluginfinder.get_plugins():
            yield from pluginmetadata.generate()

    def run(self):
        node = nodes.Element(document=self.state.document)
        result = StringList(list(self.generate()), "plugins")
        nested_parse_with_titles(self.state, result, node)

        return node.children


def setup(app):
    app.add_directive("plugins", PluginsDirective)



from streamlink.plugins.wwenetwork import WWENetwork
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlWWENetwork(PluginCanHandleUrl):
    __plugin__ = WWENetwork

    should_match_groups = [
        ("https://network.wwe.com/video/3622", {"content_type": "video", "stream_id": "3622"}),
        ("https://network.wwe.com/live/3622", {"content_type": "live", "stream_id": "3622"}),
    ]



from streamlink.plugins.hiplayer import HiPlayer
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlHiPlayer(PluginCanHandleUrl):
    __plugin__ = HiPlayer

    should_match = [
        ("alwasatly", "https://alwasat.ly/live"),
    ]



from streamlink.plugins.telemadrid import Telemadrid
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlTelemadrid(PluginCanHandleUrl):
    __plugin__ = Telemadrid

    should_match = [
        "https://www.telemadrid.es/",
        "https://www.telemadrid.es/emision-en-directo/",
        "https://www.telemadrid.es/programas/telenoticias-1/Telenoticias-1-02032023-2-2538066218--20230302042202.html",
    ]



from streamlink.plugins.nimotv import NimoTV
from tests.plugins import PluginCanHandleUrl


class TestPluginNimoTV(PluginCanHandleUrl):
    __plugin__ = NimoTV

    should_match = [
        "http://www.nimo.tv/live/737614",
        "https://www.nimo.tv/live/737614",
        "http://www.nimo.tv/sanz",
        "https://www.nimo.tv/sanz",
        "https://m.nimo.tv/user",
    ]



from __future__ import annotations

import argparse

import pytest

from streamlink.options import Argument, Arguments, Options
from streamlink.utils.args import comma_list_filter


class TestOptions:
    @pytest.fixture()
    def options(self):
        return Options({
            "foo-bar": "1",
            "baz_qux": "2",
        })

    def test_defaults(self, options: Options):
        empty = Options()
        assert empty == {}
        assert empty.defaults == {}

        assert options.defaults == {
            "foo-bar": "1",
            "baz-qux": "2",
        }
        assert options == options.defaults

    def test_get_set(self, options: Options):
        assert options.get("foo-bar") == "1"
        assert options["foo-bar"] == "1"
        assert options.get("baz_qux") == "2"
        assert options["baz_qux"] == "2"

        assert options.get("non_existing") is None
        assert options["non_existing"] is None

        options.set("abc-def", 3.14)
        assert options.get("abc-def") == 3.14  # noqa: RUF069
        assert options.get("abc-def") == 3.14  # noqa: RUF069

        obj = object()
        options["foo_bar"] = obj
        assert options.get("foo-bar") is obj
        assert options.get("foo_bar") is obj

        assert list(options.items()) == [
            ("foo-bar", obj),
            ("baz-qux", "2"),
            ("abc-def", 3.14),
        ]

    def test_update(self, options: Options):
        assert options.get("foo-bar") == "1"
        assert options.get("non_existing") is None

        options.update({"foo-bar": "value"})
        assert options.get("foo-bar") == "value"

        other = Options({"foo-bar": "VALUE"})
        other.set("abc", "def")
        options.update(other)

        assert list(options.items()) == [
            ("foo-bar", "VALUE"),
            ("baz-qux", "2"),
            ("abc", "def"),
        ]

    def test_clear(self, options: Options):
        assert options.get("foo-bar") == "1"
        options.set("foo-bar", "other")
        assert options.get("foo-bar") == "other"
        options.clear()
        assert options.get("foo-bar") == "1"


class TestMappedOptions:
    class MappedOptions(Options):
        def _get_uppercase(self, key):
            return self.get_explicit(key.upper())

        def _get_add(self, key):
            return int(self.get_explicit(key)) + 1

        def _set_uppercase(self, key, value):
            self.set_explicit(key.upper(), value)

        def _set_add(self, key, value):
            self.set_explicit(key, int(value) + 1)

        _MAP_GETTERS = {
            "foo-bar": _get_uppercase,
            "baz": _get_add,
        }

        _MAP_SETTERS = {
            "foo-bar": _set_uppercase,
            "baz": _set_add,
        }

    @pytest.fixture()
    def options(self):
        return self.MappedOptions({"foo-bar": 123, "baz": 100})

    def test_mapped_key(self, options: MappedOptions):
        assert options.get("foo-bar") is None
        assert options.get("foo_bar") is None
        assert options.get_explicit("foo-bar") == 123
        assert options.get_explicit("foo_bar") == 123
        assert options.get_explicit("FOO-BAR") is None
        assert options.get_explicit("FOO_BAR") is None

        options.set("foo-bar", 321)
        assert options.get("foo-bar") == 321
        assert options.get("foo_bar") == 321
        assert options.get_explicit("foo-bar") == 123
        assert options.get_explicit("foo_bar") == 123
        assert options.get_explicit("FOO-BAR") == 321
        assert options.get_explicit("FOO_BAR") == 321

    def test_mapped_value(self, options: MappedOptions):
        assert options.get("baz") == 101
        assert options.get_explicit("baz") == 100

        options.set("baz", 0)
        assert options.get("baz") == 2
        assert options.get_explicit("baz") == 1

    def test_mutablemapping_methods(self, options: MappedOptions):
        options["key"] = "value"
        assert options["key"] == "value"

        assert options["foo-bar"] is None

        options["baz"] = 0
        assert options["baz"] == 2

        assert "foo-bar" in options
        assert "qux" not in options

        assert len(options) == 3

        assert list(iter(options)) == ["foo-bar", "baz", "key"]
        assert list(options.keys()) == ["foo-bar", "baz", "key"]
        assert list(options.values()) == [123, 1, "value"]
        assert list(options.items()) == [("foo-bar", 123), ("baz", 1), ("key", "value")]


class TestArgument:
    def test_name(self):
        assert Argument("test-arg").argument_name("plugin") == "--plugin-test-arg"
        assert Argument("test-arg").namespace_dest("plugin") == "plugin_test_arg"
        assert Argument("test-arg").dest == "test_arg"

    def test_name_plugin(self):
        assert Argument("test-arg").argument_name("test_plugin") == "--test-plugin-test-arg"
        assert Argument("test-arg").namespace_dest("test_plugin") == "test_plugin_test_arg"
        assert Argument("test-arg").dest == "test_arg"

    def test_name_override(self):
        assert Argument("test", argument_name="override-name").argument_name("plugin") == "--override-name"
        assert Argument("test", argument_name="override-name").namespace_dest("plugin") == "override_name"
        assert Argument("test", argument_name="override-name").dest == "test"

    def test_default(self):
        arg = Argument("test", default=123)
        assert arg.default == 123
        with pytest.raises(AttributeError):
            # noinspection PyPropertyAccess
            arg.default = 456  # type: ignore

    def test_help_suppress(self):
        argA = Argument("test", help="test")
        argB = Argument("test", help=argparse.SUPPRESS)
        argC = Argument("test", help=f"{argparse.SUPPRESS[:1]}{argparse.SUPPRESS[1:]}")
        assert argA.help is not argparse.SUPPRESS
        assert id(argB.help) == id(argparse.SUPPRESS)
        assert id(argC.help) == id(argparse.SUPPRESS)

    def test_options(self):
        arg = Argument(
            "test",
            action="append",
            nargs=2,
            default=(0, 0),
            type=int,
            choices=[1, 2, 3],
            required=True,
            help=argparse.SUPPRESS,
            metavar=("ONE", "TWO"),
            dest="dest",
            requires=["other"],
            prompt="Test!",
            sensitive=False,
            argument_name=None,
        )
        assert arg.options == {
            "action": "append",
            "nargs": 2,
            "default": (0, 0),
            "type": int,
            "choices": (1, 2, 3),
            "help": argparse.SUPPRESS,
            "metavar": ("ONE", "TWO"),
            "dest": "dest",
        }

        arg = Argument(
            "test",
            action="store_const",
            const=123,
        )
        assert arg.options == {
            "action": "store_const",
            "const": 123,
        }

        # doesn't include the const keyword if action is store_true or store_false
        assert Argument("test", action="store_true").options == {"action": "store_true", "default": False}
        assert Argument("test", action="store_true", default=123).options == {"action": "store_true", "default": 123}
        assert Argument("test", action="store_false").options == {"action": "store_false", "default": True}
        assert Argument("test", action="store_false", default=123).options == {"action": "store_false", "default": 123}

    def test_equality(self):
        a1 = Argument(
            "test",
            action="append",
            nargs=2,
            default=("0", "0"),
            type=comma_list_filter(["1", "2", "3"], unique=True),
            choices=["1", "2", "3"],
            required=True,
            help=argparse.SUPPRESS,
            metavar=("ONE", "TWO"),
            dest="dest",
            requires=["other"],
            prompt="Test!",
            sensitive=False,
            argument_name="custom-name",
        )
        a2 = Argument(
            "test",
            action="append",
            nargs=2,
            default=("0", "0"),
            type=comma_list_filter(["1", "2", "3"], unique=True),
            choices=["1", "2", "3"],
            required=True,
            help=argparse.SUPPRESS,
            metavar=("ONE", "TWO"),
            dest="dest",
            requires=["other"],
            prompt="Test!",
            sensitive=False,
            argument_name="custom-name",
        )
        a3 = Argument("foo")
        assert a1 is not a2
        assert a1 == a2
        assert a1 != a3
        assert a2 != a3

    @pytest.mark.parametrize(
        ("action", "default", "expected_const", "expected_default"),
        [
            pytest.param("store_true", None, True, False, id="store_true-no-default"),
            pytest.param("store_true", "default", True, "default", id="store_true-with-default"),
            pytest.param("store_false", None, False, True, id="store_false-no-default"),
            pytest.param("store_false", "default", False, "default", id="store_false-with-default"),
        ],
    )
    def test_store_true_false(self, action: str, default: str | None, expected_const: bool, expected_default: str | None):
        arg = Argument("foo", action=action, default=default)
        assert arg.const is expected_const
        assert arg.default is expected_default


class TestArguments:
    def test_getter(self):
        test1 = Argument("test1")
        test2 = Argument("test2")
        args = Arguments(test1, test2)

        assert args.get("test1") == test1
        assert args.get("test2") == test2
        assert args.get("test3") is None

    def test_iter(self):
        test1 = Argument("test1")
        test2 = Argument("test2")
        args = Arguments(test1, test2)

        i_args = iter(args)
        assert next(i_args) == test1
        assert next(i_args) == test2

    def test_add(self):
        test1 = Argument("test1")
        test2 = Argument("test2")
        test3 = Argument("test3")
        args = Arguments(test1, test2)

        assert list(iter(args)) == [test1, test2]
        args.add(test3)
        assert list(iter(args)) == [test3, test1, test2]

    def test_equality(self):
        test1 = Arguments()
        test1.add(Argument("testA"))
        test1.add(Argument("testB"))
        test2 = Arguments(Argument("testB"), Argument("testA"))
        test3 = Arguments(Argument("testA"), Argument("testB"))
        assert test1 == test2
        assert test1 != test3
        assert test2 != test3

    def test_requires(self):
        test1 = Argument("test1", requires="test2")
        test2 = Argument("test2", requires="test3")
        test3 = Argument("test3")
        args = Arguments(test1, test2, test3)

        assert list(args.requires("test1")) == [test2, test3]

    def test_requires_invalid(self):
        test1 = Argument("test1", requires="test2")
        args = Arguments(test1)

        with pytest.raises(KeyError) as cm:
            list(args.requires("test1"))
        assert cm.value.args[0] == "test2 is not a valid argument for this plugin"

    @pytest.mark.parametrize(
        "args",
        [
            pytest.param(
                Arguments(
                    Argument("test1", requires="test2"),
                    Argument("test2", requires="test1"),
                ),
                id="Cycle",
            ),
            pytest.param(
                Arguments(
                    Argument("test1", requires="test2"),
                    Argument("test2", requires="test3"),
                    Argument("test3", requires="test1"),
                ),
                id="Cycle deep",
            ),
            pytest.param(
                Arguments(
                    Argument("test1", requires="test1"),
                ),
                id="Cycle self",
            ),
        ],
    )
    def test_requires_cycle(self, args: Arguments):
        with pytest.raises(RuntimeError) as cm:
            list(args.requires("test1"))
        assert cm.value.args[0] == "cycle detected in plugin argument config"



from __future__ import annotations

import logging
from pathlib import Path

from setuptools import Command
from setuptools.command.build_py import build_py

from build_backend.plugins_json import build, to_json


STREAMLINK_PLUGINS_JSON = Path("streamlink", "plugins", "_plugins.json")


__all__ = ["cmdclass"]


class StreamlinkBuildPyCommand(build_py, Command):
    def _build_plugins_json(self) -> None:
        self.announce("building plugins JSON data", logging.INFO)
        output = Path(self.build_lib) / STREAMLINK_PLUGINS_JSON
        output.parent.mkdir(parents=True, exist_ok=True)
        data = build()
        with output.open("w", encoding="utf-8") as fd:
            to_json(data, fd)

    def build_package_data(self) -> None:
        super().build_package_data()

        self._build_plugins_json()


cmdclass: dict[str, type[Command]] = {
    "build_py": StreamlinkBuildPyCommand,
}



import pytest

from streamlink.stream.segmented.segment import Segment
from streamlink.stream.segmented.segmented import log


def test_logger_name():
    assert log.name == "streamlink.stream.segmented"


@pytest.mark.parametrize(
    ("data", "expected"),
    [
        pytest.param(
            dict(
                num=1,
                init=False,
                discontinuity=False,
                uri="/path/to/segment.ts?query#fragment",
                duration=4.0,
            ),
            "Segment(num=1, init=False, discontinuity=False, duration=4.000, fileext='ts')",
            id="with-fileext",
        ),
        pytest.param(
            dict(
                num=1,
                init=False,
                discontinuity=False,
                uri="/path/to/segment.other?query#fragment",
                duration=4.0,
            ),
            "Segment(num=1, init=False, discontinuity=False, duration=4.000, fileext=None)",
            id="without-fileext",
        ),
    ],
)
def test_segment_serialization(data: dict, expected: str):
    segment = Segment(**data)
    assert repr(segment) == expected



from streamlink.plugins.nos import NOS
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlNOS(PluginCanHandleUrl):
    __plugin__ = NOS

    should_match = [
        "https://nos.nl/live",
        "https://nos.nl/livestream/2539164-schaatsen-wb-milwaukee-straks-de-massastart-m",
        "https://nos.nl/collectie/13951/video/2491092-dit-was-prinsjesdag",
        "https://nos.nl/l/2490788",
        "https://nos.nl/video/2490788-meteoor-gespot-boven-noord-nederland",
    ]

    should_not_match = [
        "https://nos.nl/livestream",
        "https://nos.nl/l",
        "https://nos.nl/video",
        "https://nos.nl/collectie",
        "https://nos.nl/artikel/2385784-explosieve-situatie-leidde-tot-verwoeste-huizen-en-omgewaaide-bomen-leersum",
        "https://nos.nl/sport",
        "https://nos.nl/sport/videos",
        "https://nos.nl/programmas",
        "https://nos.nl/uitzendingen",
        "https://nos.nl/uitzendingen/livestream/2385462",
    ]



from streamlink.plugins.blasttv import BlastTv
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlBlastTv(PluginCanHandleUrl):
    __plugin__ = BlastTv

    should_match_groups = [
        # live
        (("live", "https://blast.tv"), {}),
        (("live", "https://blast.tv/"), {}),
        (("live", "https://blast.tv/live"), {}),
        (("live", "https://blast.tv/live/"), {}),
        (("live", "https://blast.tv/live/a"), {"channel": "a"}),
        (("live", "https://blast.tv/live/co-stream"), {"channel": "co-stream"}),
        (("live", "https://blast.tv/live/americas1"), {"channel": "americas1"}),
        # external streams (at the time of testing)
        (("live", "https://blast.tv/live/f"), {"channel": "f"}),
        # VODs
        (
            ("vod", "https://blast.tv/cs/tournaments/rivals-2025-season-1/match/bfaaa42e/falcons-vitality"),
            {"game": "cs", "shortid": "bfaaa42e"},
        ),
        (
            ("vod", "https://blast.tv/cs/tournaments/open-2025-season-1/match/24ff9d6c/vitality-mouz"),
            {"game": "cs", "shortid": "24ff9d6c"},
        ),
        # external VODs
        (
            ("vod", "https://blast.tv/dota/tournaments/blast-slam-iii/match/acabb915/falcons-tundra"),
            {"game": "dota", "shortid": "acabb915"},
        ),
    ]

    should_not_match = [
        "https://blast.tv/a",
    ]



from __future__ import annotations

import importlib.util
from io import BufferedWriter
from os import devnull
from typing import TYPE_CHECKING
from unittest.mock import Mock, call

import streamlink_cli.compat


if TYPE_CHECKING:
    import pytest


def test_no_stdout(monkeypatch: pytest.MonkeyPatch):
    monkeypatch.setattr("sys.stdout", None)

    mock_atexit_register = Mock()
    monkeypatch.setattr("atexit.register", mock_atexit_register)

    spec = importlib.util.find_spec("streamlink_cli.compat")
    assert spec
    assert spec.loader

    module = importlib.util.module_from_spec(spec)
    assert module is not streamlink_cli.compat

    spec.loader.exec_module(module)
    assert module.sys.stdout is None
    assert isinstance(module.stdout, BufferedWriter)
    assert module.stdout.name == devnull

    assert mock_atexit_register.call_args_list == [call(module.stdout.close)]
    module.stdout.close()  # close manually, since we've mocked the atexit.register() call



from streamlink.plugins.raiplay import RaiPlay
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlRaiPlay(PluginCanHandleUrl):
    __plugin__ = RaiPlay

    should_match = [
        ("live", "https://raiplay.it/dirette/rai1"),
        ("live", "https://www.raiplay.it/dirette/rai1"),
        ("live", "https://www.raiplay.it/dirette/rai2"),
        ("live", "https://www.raiplay.it/dirette/rai3"),
        ("live", "https://www.raiplay.it/dirette/rainews24"),
        (
            "vod",
            "https://raiplay.it/video/2023/11/Un-posto-al-sole---Puntata-del-08112023-EP6313-1377bcf9-db3f-40f7-aa05-fefeb086ec68.html",
        ),
        (
            "vod",
            "https://www.raiplay.it/video/2023/11/Un-posto-al-sole---Puntata-del-08112023-EP6313-1377bcf9-db3f-40f7-aa05-fefeb086ec68.html",
        ),
    ]



from __future__ import annotations

from io import BytesIO, TextIOWrapper
from os import linesep
from typing import TYPE_CHECKING
from unittest.mock import Mock

import pytest

from streamlink_cli.console.stream import (
    ConsoleOutputStream,
    ConsoleOutputStreamANSI,
    ConsoleOutputStreamWindows,
    ConsoleStatusMessage,
)


if TYPE_CHECKING:
    from typing_extensions import Buffer


nl = linesep.encode("ascii")


@pytest.fixture()
def buffer():
    return BytesIO()


@pytest.fixture()
def stream(buffer: BytesIO):
    # stdout/stderr is line buffered, which implies flushes when writing \n or \r
    return TextIOWrapper(buffer, encoding="utf-8", line_buffering=True)


@pytest.fixture()
def console_output_stream(stream: TextIOWrapper):
    return ConsoleOutputStream(stream)


class TestConsoleOutputStreamFeatureDetection:
    @pytest.fixture(autouse=True)
    def _mock_is_win32(self, request: pytest.FixtureRequest, monkeypatch: pytest.MonkeyPatch):
        monkeypatch.setattr("streamlink_cli.console.stream.is_win32", getattr(request, "param", False))

    @pytest.fixture(autouse=True)
    def _mock_windows_console(self, request: pytest.FixtureRequest, monkeypatch: pytest.MonkeyPatch):
        value = getattr(request, "param", None)
        mock_windows_console = Mock(supports_virtual_terminal_processing=Mock(return_value=value))
        MockWindowsConsole = Mock(return_value=mock_windows_console if value is not None else None)
        monkeypatch.setattr("streamlink_cli.console.stream.WindowsConsole", MockWindowsConsole)

    @pytest.fixture()
    def stream(self, request: pytest.FixtureRequest, monkeypatch: pytest.MonkeyPatch, stream: TextIOWrapper):
        isatty = getattr(request, "param", True)
        setattr(stream, "isatty", lambda: isatty)  # noqa: B010

        return stream

    # noinspection PyTestParametrized
    @pytest.mark.parametrize(
        ("_mock_is_win32", "stream", "os_environ", "_mock_windows_console", "expected"),
        [
            pytest.param(False, False, {}, None, ConsoleOutputStream, id="posix-notatty"),
            pytest.param(False, True, {}, None, ConsoleOutputStreamANSI, id="posix-isatty"),
            pytest.param(False, True, {"TERM": "dumb"}, None, ConsoleOutputStream, id="posix-TERM=dumb"),
            pytest.param(False, True, {"TERM": "unknown"}, None, ConsoleOutputStream, id="posix-TERM=unknown"),
            pytest.param(True, False, {}, None, ConsoleOutputStream, id="win32-notatty"),
            pytest.param(True, True, {}, None, ConsoleOutputStreamANSI, id="win32-nowindowsconsole"),
            pytest.param(True, True, {}, True, ConsoleOutputStreamANSI, id="win32-windowsconsole-virtprocessing"),
            pytest.param(True, True, {}, False, ConsoleOutputStreamWindows, id="win32-windowsconsole-novirtprocessing"),
        ],
        indirect=["_mock_is_win32", "os_environ", "stream", "_mock_windows_console"],
    )
    def test_class_type(self, os_environ: dict, console_output_stream: ConsoleOutputStream, expected: ConsoleOutputStream):
        assert type(console_output_stream) is expected


class TestConsoleOutputStream:
    @pytest.fixture(autouse=True)
    def _mock_new(self, monkeypatch: pytest.MonkeyPatch):
        monkeypatch.setattr(ConsoleOutputStream, "__new__", lambda *_, **__: object.__new__(ConsoleOutputStream))

    def test_supports_status_messages(self, console_output_stream: ConsoleOutputStream):
        assert not console_output_stream.supports_status_messages()

    def test_write(self, buffer: BytesIO, console_output_stream: ConsoleOutputStream):
        assert buffer.getvalue() == b""

        assert console_output_stream.write("foo") == 0
        assert buffer.getvalue() == b""
        assert console_output_stream._line_buffer == ["foo"]

        assert console_output_stream.write("bar") == 0
        assert buffer.getvalue() == b""
        assert console_output_stream._line_buffer == ["foo", "bar"]

        assert console_output_stream.write("baz\n123\n456\n") == (6 + 4) + 4 + 4
        assert buffer.getvalue() == b"foobarbaz" + nl + b"123" + nl + b"456" + nl
        assert console_output_stream._line_buffer == []

        assert console_output_stream.write("abc") == 0
        console_output_stream.flush()
        assert buffer.getvalue() == b"foobarbaz" + nl + b"123" + nl + b"456" + nl + b"abc"
        assert console_output_stream._line_buffer == []

        console_output_stream.writelines(["QWERTY\n", "ASDF"])
        assert buffer.getvalue() == b"foobarbaz" + nl + b"123" + nl + b"456" + nl + b"abcQWERTY" + nl
        assert console_output_stream._line_buffer == ["ASDF"]
        console_output_stream.flush()
        assert buffer.getvalue() == b"foobarbaz" + nl + b"123" + nl + b"456" + nl + b"abcQWERTY" + nl + b"ASDF"

    def test_write_status_message(self, buffer: BytesIO, console_output_stream: ConsoleOutputStream):
        assert console_output_stream.write(ConsoleStatusMessage("foo")) == 0
        assert console_output_stream._line_buffer == []
        console_output_stream.flush()
        assert buffer.getvalue() == b""

    def test_close(self, buffer: BytesIO, console_output_stream: ConsoleOutputStream):
        def fakewrite(s: Buffer) -> int:
            msg = bytes(s)
            fakebuffer.append(msg)

            return len(msg)

        fakebuffer: list[Buffer] = []
        buffer.write = fakewrite  # type: ignore[method-assign, ty:invalid-assignment]

        assert console_output_stream.write("foo") == 0
        assert fakebuffer == []
        assert console_output_stream._line_buffer == ["foo"]
        assert not console_output_stream.closed

        console_output_stream.close()
        assert fakebuffer == [b"foo"]
        assert console_output_stream._line_buffer == []
        assert console_output_stream.closed

        console_output_stream.close()  # noop - does not raise

        with pytest.raises(ValueError, match=r"^I/O operation on closed file\.$"):
            console_output_stream.write("foo")

        with pytest.raises(ValueError, match=r"^I/O operation on closed file\.$"):
            console_output_stream.writelines(["foo", "bar"])

        with pytest.raises(ValueError, match=r"^I/O operation on closed file\.$"):
            console_output_stream.flush()


class TestConsoleOutputStreamANSI:
    @pytest.fixture(autouse=True)
    def _mock_new(self, monkeypatch: pytest.MonkeyPatch):
        monkeypatch.setattr(ConsoleOutputStream, "__new__", lambda *_, **__: object.__new__(ConsoleOutputStreamANSI))

    @pytest.fixture()
    def fakebuffer(self, buffer: BytesIO):
        def fakewrite(s: Buffer) -> int:
            msg = bytes(s)
            fakebuffer.append(msg)

            return len(msg)

        fakebuffer: list[Buffer] = []
        buffer.write = fakewrite  # type: ignore[method-assign, ty:invalid-assignment]

        return fakebuffer

    def test_supports_status_messages(self, console_output_stream: ConsoleOutputStreamANSI):
        assert console_output_stream.supports_status_messages()

    def test_write(self, buffer: BytesIO, console_output_stream: ConsoleOutputStreamANSI):
        clr_eol = console_output_stream._CR_CLREOL.encode("ascii")
        len_clr_eol = len(clr_eol)
        assert len_clr_eol == 4

        assert console_output_stream.write("foo") == 0
        assert buffer.getvalue() == b""
        assert console_output_stream._line_buffer == ["foo"]
        assert console_output_stream._last_status == ""

        assert console_output_stream.write("bar\n") == 3 + 4
        assert buffer.getvalue() == b"foobar" + nl
        assert console_output_stream._line_buffer == []
        assert console_output_stream._last_status == ""

        assert console_output_stream.write(ConsoleStatusMessage("123\n")) == 3
        assert buffer.getvalue() == b"foobar" + nl + b"123"
        assert console_output_stream._line_buffer == []
        assert console_output_stream._last_status == "123"

        assert console_output_stream.write(ConsoleStatusMessage("456\n")) == len_clr_eol + 3
        assert buffer.getvalue() == b"foobar" + nl + b"123" + clr_eol + b"456"
        assert console_output_stream._line_buffer == []
        assert console_output_stream._last_status == "456"

        assert console_output_stream.write("abc") == 0
        assert buffer.getvalue() == b"foobar" + nl + b"123" + clr_eol + b"456"
        assert console_output_stream._line_buffer == ["abc"]
        assert console_output_stream._last_status == "456"

        assert console_output_stream.write(ConsoleStatusMessage("789\n")) == len_clr_eol + 3
        assert buffer.getvalue() == b"foobar" + nl + b"123" + clr_eol + b"456" + clr_eol + b"789"
        assert console_output_stream._line_buffer == ["abc"]
        assert console_output_stream._last_status == "789"

        assert console_output_stream.write("def\n") == 3 + 4 + len_clr_eol + len(console_output_stream._last_status)
        assert buffer.getvalue() == (
            b"foobar" + nl + b"123" + clr_eol + b"456" + clr_eol + b"789" + clr_eol + b"abcdef" + nl + b"789"
        )
        assert console_output_stream._line_buffer == []
        assert console_output_stream._last_status == "789"

    def test_close(self, console_output_stream: ConsoleOutputStreamANSI):
        console_output_stream.close()

        with pytest.raises(ValueError, match=r"^I/O operation on closed file\.$"):
            console_output_stream.write("foo")

        with pytest.raises(ValueError, match=r"^I/O operation on closed file\.$"):
            console_output_stream.writelines(["foo", "bar"])

        with pytest.raises(ValueError, match=r"^I/O operation on closed file\.$"):
            console_output_stream.flush()

    def test_close_without_line_buffer_without_status(
        self,
        fakebuffer: list[bytes],
        console_output_stream: ConsoleOutputStreamANSI,
    ):
        assert console_output_stream.write("foo\n") == 4
        assert fakebuffer == [b"foo" + nl]
        assert console_output_stream._line_buffer == []
        assert console_output_stream._last_status == ""
        assert not console_output_stream.closed

        console_output_stream.close()
        assert fakebuffer == [b"foo" + nl]
        assert console_output_stream._line_buffer == []
        assert console_output_stream._last_status == ""
        assert console_output_stream.closed

        console_output_stream.close()  # noop - does not raise

    def test_close_with_line_buffer_without_status(
        self,
        fakebuffer: list[bytes],
        console_output_stream: ConsoleOutputStreamANSI,
    ):
        assert console_output_stream.write("foo\n") == 4
        assert fakebuffer == [b"foo" + nl]
        assert console_output_stream._line_buffer == []
        assert console_output_stream._last_status == ""
        assert not console_output_stream.closed

        assert console_output_stream.write("bar") == 0
        assert fakebuffer == [b"foo" + nl]
        assert console_output_stream._line_buffer == ["bar"]
        assert console_output_stream._last_status == ""
        assert not console_output_stream.closed

        console_output_stream.close()
        assert fakebuffer == [b"foo" + nl, b"bar" + nl]
        assert console_output_stream._line_buffer == []
        assert console_output_stream._last_status == ""
        assert console_output_stream.closed

        console_output_stream.close()  # noop - does not raise

    def test_close_without_line_buffer_with_status(
        self,
        fakebuffer: list[bytes],
        console_output_stream: ConsoleOutputStreamANSI,
    ):
        assert console_output_stream.write("foo\n") == 4
        assert fakebuffer == [b"foo" + nl]
        assert console_output_stream._line_buffer == []
        assert console_output_stream._last_status == ""
        assert not console_output_stream.closed

        assert console_output_stream.write(ConsoleStatusMessage("123\n")) == 3
        assert fakebuffer == [b"foo" + nl, b"123"]
        assert console_output_stream._line_buffer == []
        assert console_output_stream._last_status == "123"
        assert not console_output_stream.closed

        console_output_stream.close()
        assert fakebuffer == [b"foo" + nl, b"123", nl]
        assert console_output_stream._line_buffer == []
        assert console_output_stream._last_status == ""
        assert console_output_stream.closed

        console_output_stream.close()  # noop - does not raise

    def test_close_with_line_buffer_with_status(
        self,
        fakebuffer: list[bytes],
        console_output_stream: ConsoleOutputStreamANSI,
    ):
        clr_eol = console_output_stream._CR_CLREOL.encode("ascii")

        assert console_output_stream.write("foo\n") == 4
        assert fakebuffer == [b"foo" + nl]
        assert console_output_stream._line_buffer == []
        assert console_output_stream._last_status == ""
        assert not console_output_stream.closed

        assert console_output_stream.write("bar") == 0
        assert fakebuffer == [b"foo" + nl]
        assert console_output_stream._line_buffer == ["bar"]
        assert console_output_stream._last_status == ""
        assert not console_output_stream.closed

        assert console_output_stream.write(ConsoleStatusMessage("123\n")) == 3
        assert fakebuffer == [b"foo" + nl, b"123"]
        assert console_output_stream._line_buffer == ["bar"]
        assert console_output_stream._last_status == "123"
        assert not console_output_stream.closed

        console_output_stream.close()
        assert fakebuffer == [b"foo" + nl, b"123", clr_eol + b"bar" + nl + b"123" + nl]
        assert console_output_stream._line_buffer == []
        assert console_output_stream._last_status == ""
        assert console_output_stream.closed

        console_output_stream.close()  # noop - does not raise


class TestConsoleOutputStreamWindows:
    @pytest.fixture(autouse=True)
    def _mock_new(self, monkeypatch: pytest.MonkeyPatch):
        monkeypatch.setattr(ConsoleOutputStream, "__new__", lambda *_, **__: object.__new__(ConsoleOutputStreamWindows))

    @pytest.fixture(autouse=True)
    def mock_windows_console(self, monkeypatch: pytest.MonkeyPatch):
        mock_windows_console = Mock()
        monkeypatch.setattr(ConsoleOutputStreamWindows, "windows_console", mock_windows_console, raising=False)

        return mock_windows_console

    def test_supports_status_messages(self) -> None:
        writer = ConsoleOutputStream(TextIOWrapper(BytesIO(), encoding="utf-8"))
        assert writer.supports_status_messages()

    def test_write(self, buffer: BytesIO, console_output_stream: ConsoleOutputStreamWindows, mock_windows_console: Mock):
        assert console_output_stream.write("foo") == 0
        assert buffer.getvalue() == b""
        assert console_output_stream._line_buffer == ["foo"]
        assert console_output_stream._last_status == ""
        assert mock_windows_console.clear_line.call_count == 0

        assert console_output_stream.write("bar\n") == 3 + 4
        assert buffer.getvalue() == b"foobar" + nl
        assert console_output_stream._line_buffer == []
        assert console_output_stream._last_status == ""
        assert mock_windows_console.clear_line.call_count == 0

        assert console_output_stream.write(ConsoleStatusMessage("123\n")) == 3
        assert buffer.getvalue() == b"foobar" + nl + b"123"
        assert console_output_stream._line_buffer == []
        assert console_output_stream._last_status == "123"
        assert mock_windows_console.clear_line.call_count == 0

        assert console_output_stream.write(ConsoleStatusMessage("456\n")) == 3
        assert buffer.getvalue() == b"foobar" + nl + b"123456"
        assert console_output_stream._line_buffer == []
        assert console_output_stream._last_status == "456"
        assert mock_windows_console.clear_line.call_count == 1

        assert console_output_stream.write("abc") == 0
        assert buffer.getvalue() == b"foobar" + nl + b"123456"
        assert console_output_stream._line_buffer == ["abc"]
        assert console_output_stream._last_status == "456"
        assert mock_windows_console.clear_line.call_count == 1

        assert console_output_stream.write(ConsoleStatusMessage("789\n")) == 3
        assert buffer.getvalue() == b"foobar" + nl + b"123456789"
        assert console_output_stream._line_buffer == ["abc"]
        assert console_output_stream._last_status == "789"
        assert mock_windows_console.clear_line.call_count == 2

        assert console_output_stream.write("def\n") == 3 + 4 + len(console_output_stream._last_status)
        assert buffer.getvalue() == b"foobar" + nl + b"123456789abcdef" + nl + b"789"
        assert console_output_stream._line_buffer == []
        assert console_output_stream._last_status == "789"
        assert mock_windows_console.clear_line.call_count == 3



from __future__ import annotations

import logging
import warnings
from datetime import timezone
from errno import EINVAL, EPIPE
from inspect import currentframe, getframeinfo
from io import BytesIO, TextIOWrapper
from pathlib import Path
from typing import TYPE_CHECKING

import freezegun
import pytest

from streamlink import logger
from streamlink.exceptions import StreamlinkDeprecationWarning, StreamlinkWarning
from streamlink.logger import StreamlinkLogger, getLogger


if TYPE_CHECKING:
    from collections.abc import Iterable


def getvalue(output: TextIOWrapper, size: int = -1):
    output.seek(0)

    return output.read(size)


@pytest.fixture()
def output(request: pytest.FixtureRequest):
    params = getattr(request, "param", {})
    params.setdefault("encoding", "utf-8")
    output = TextIOWrapper(BytesIO(), **params)

    return output


@pytest.fixture()
def logfile(tmp_path: Path) -> Path:
    return tmp_path / "log.txt"


@pytest.fixture()
def log(request: pytest.FixtureRequest, monkeypatch: pytest.MonkeyPatch, output: TextIOWrapper):
    params = getattr(request, "param", {})
    params.setdefault("format", "[{name}][{levelname}] {message}")
    params.setdefault("style", "{")

    if "logfile" in request.fixturenames:
        params["filename"] = request.getfixturevalue("logfile")

    fakeroot = getLogger("streamlink.test")

    monkeypatch.setattr("streamlink.logger.root", fakeroot)
    monkeypatch.setattr("streamlink.utils.times.LOCAL", timezone.utc)

    stream: TextIOWrapper | None = output if params.pop("stream", True) else None
    handler = logger.basicConfig(stream=stream, **params)
    if stream is None:
        assert handler is None
    else:
        assert isinstance(handler, logging.StreamHandler)

    yield fakeroot

    logger.capturewarnings(False)

    if handler:
        handler.close()
        fakeroot.removeHandler(handler)
    assert not fakeroot.handlers


class TestLogging:
    @pytest.fixture()
    def log_failure(
        self,
        request: pytest.FixtureRequest,
        monkeypatch: pytest.MonkeyPatch,
        log: logging.Logger,
        output: TextIOWrapper,
    ):
        params = getattr(request, "param", {})

        with pytest.raises(Exception) as cm:  # noqa: PT011
            logger.basicConfig(stream=output, **params)

        return cm.value

    def test_root(self):
        assert isinstance(logger.root, StreamlinkLogger)
        assert logger.root.name == "streamlink"
        assert logger.root.parent is logging.root

    def test_get_logger(self):
        child_a = getLogger("streamlink.test")
        child_b = getLogger("test")
        child_c = getLogger("other")
        assert isinstance(child_a, StreamlinkLogger)
        assert child_b is child_a
        assert child_a.parent is logger.root
        assert isinstance(child_c, StreamlinkLogger)
        assert child_c is not child_a
        assert child_c.parent is logger.root

    def test_logging_get_logger(self):
        root = logging.getLogger("streamlink")
        assert isinstance(root, StreamlinkLogger)
        assert root is logger.root
        # we can't safely make an assertion about the logger class when calling `logging.getLogger()`,
        # as it depends on the manager's loggerClass value or the last `logging.setLoggerClass()` call.
        child = logging.getLogger("streamlink.doesnotexistyet")
        assert child.parent is logger.root
        assert logging.getLogger("other").parent is not logger.root

    @pytest.mark.parametrize(
        ("name", "level"),
        [
            ("none", logger.NONE),
            ("critical", logger.CRITICAL),
            ("error", logger.ERROR),
            ("warning", logger.WARNING),
            ("info", logger.INFO),
            ("debug", logger.DEBUG),
            ("trace", logger.TRACE),
            ("all", logger.ALL),
        ],
    )
    def test_level_names(self, name: str, level: int):
        assert logging.getLevelName(level) == name
        assert logging.getLevelName(name) == level
        assert logging.getLevelName(name.upper()) == level

    def test_levels_list(self):
        assert logger.levels == ["none", "critical", "error", "warning", "info", "debug", "trace", "all"]

    def test_default_level(self):
        assert logger.root.level == logger.WARNING

    def test_level(self, log: StreamlinkLogger, output: TextIOWrapper):
        log.setLevel("info")
        log.debug("test")
        output.seek(0)
        assert output.tell() == 0

        log.setLevel("debug")
        log.debug("test")
        assert output.tell() != 0

    def test_level_none(self, log: StreamlinkLogger, output: TextIOWrapper):
        log.setLevel("none")
        log.critical("test")
        log.error("test")
        log.warning("test")
        log.info("test")
        log.debug("test")
        log.trace("test")
        assert not getvalue(output)

    def test_output(self, log: StreamlinkLogger, output: TextIOWrapper):
        log.setLevel("debug")
        log.debug("test")
        assert getvalue(output) == "[test][debug] test\n"

    @pytest.mark.parametrize(
        ("log", "level", "expected"),
        [
            ({"level": logger.DEBUG}, logger.TRACE, ""),
            ({"level": logger.TRACE}, logger.TRACE, "[test][trace] test\n"),
            ({"level": logger.TRACE}, logger.DEBUG, "[test][debug] test\n"),
            ({"level": logger.TRACE}, logger.ALL, ""),
            ({"level": logger.ALL}, logger.ALL, "[test][all] test\n"),
            ({"level": logger.ALL}, logger.TRACE, "[test][trace] test\n"),
        ],
        indirect=["log"],
    )
    def test_custom_output(self, log: StreamlinkLogger, output: TextIOWrapper, level: int, expected: str):
        log.log(level, "test")
        assert getvalue(output) == expected

    # https://github.com/streamlink/streamlink/issues/4862
    @pytest.mark.parametrize(
        ("level", "levelname"),
        [
            (logger.TRACE, "trace"),
            (logger.ALL, "all"),
        ],
    )
    def test_custom_module_name(self, caplog: pytest.LogCaptureFixture, log: StreamlinkLogger, level: int, levelname: str):
        caplog.set_level(1)
        log = getLogger(self.__class__.__module__)
        getattr(log, levelname)("foo")
        log.log(level, "bar")
        assert [(record.module, record.levelname, record.message) for record in caplog.records] == [
            ("test_logger", levelname, "foo"),
            ("test_logger", levelname, "bar"),
        ]

    def test_logrecord(self, caplog: pytest.LogCaptureFixture, log: StreamlinkLogger):
        def get_lineno():
            frame = currentframe()
            assert frame
            assert frame.f_back
            return getframeinfo(frame.f_back).lineno

        caplog.set_level(1, "streamlink")
        log.info("test %(a)s %(b)s", {"a": "foo", "b": "bar"}, exc_info=ValueError("err"), stack_info=True)
        lineno = get_lineno() - 1

        assert [
            (
                r.exc_text,
                r.filename,
                r.funcName,
                r.levelname,
                r.levelno,
                r.lineno,
                r.message,
                r.module,
                r.name,
                r.pathname,
                r.stack_info is not None and len(r.stack_info) > 0,
            )
            for r in caplog.records
        ] == [
            (
                "ValueError: err",
                "test_logger.py",
                "test_logrecord",
                "info",
                20,
                lineno,
                "test foo bar",
                "test_logger",
                "test",
                __file__,
                True,
            ),
        ]

    @pytest.mark.parametrize(
        ("level", "expected"),
        [
            (logger.DEBUG, ""),
            (logger.INFO, "[test][info] foo\n[test][info] bar\n"),
        ],
    )
    def test_iter(self, log: StreamlinkLogger, output: TextIOWrapper, level: int, expected: str):
        def iterator():
            yield "foo"
            yield "bar"

        log.setLevel(logger.INFO)
        assert list(log.iter(level, iterator())) == ["foo", "bar"]
        assert getvalue(output) == expected

    @pytest.mark.parametrize(
        "log",
        [
            {"style": "%", "format": "[%(name)s][%(levelname)s] %(message)s"},
        ],
        indirect=True,
    )
    def test_style_percent(self, log: StreamlinkLogger, output: TextIOWrapper):
        log.setLevel("info")
        log.info("test")
        assert getvalue(output) == "[test][info] test\n"

    @pytest.mark.parametrize("log_failure", [{"style": "invalid"}], indirect=True)
    def test_style_invalid(self, log_failure):
        assert type(log_failure) is ValueError
        assert str(log_failure) == "Style must be one of: %,{,$"

    @freezegun.freeze_time("2000-01-02T03:04:05.123456Z")
    @pytest.mark.parametrize(
        "log",
        [
            {"format": "[{asctime}][{name}][{levelname}] {message}"},
        ],
        indirect=True,
    )
    def test_datefmt_default(self, log: StreamlinkLogger, output: TextIOWrapper):
        log.setLevel("info")
        log.info("test")
        assert getvalue(output) == "[03:04:05][test][info] test\n"

    @freezegun.freeze_time("2000-01-02T03:04:05.123456Z")
    @pytest.mark.parametrize(
        "log",
        [
            {"format": "[{asctime}][{name}][{levelname}] {message}", "datefmt": "%H:%M:%S.%f%z"},
        ],
        indirect=True,
    )
    def test_datefmt_custom(self, log: StreamlinkLogger, output: TextIOWrapper):
        log.setLevel("info")
        log.info("test")
        assert getvalue(output) == "[03:04:05.123456+0000][test][info] test\n"

    @pytest.mark.parametrize(
        ("output", "expected"),
        [
            pytest.param(
                {"encoding": "utf-8"},
                "Bär: 🐻",
                id="utf-8",
            ),
            pytest.param(
                {"encoding": "ascii"},
                "B\\xe4r: \\U0001f43b",
                id="ascii",
            ),
            pytest.param(
                {"encoding": "iso-8859-1"},
                "Bär: \\U0001f43b",
                id="iso-8859-1",
            ),
            pytest.param(
                {"encoding": "shift_jis"},
                "B\\xe4r: \\U0001f43b",
                id="shift_jis",
            ),
        ],
        indirect=["output"],
    )
    def test_logstream_encoding(self, log: StreamlinkLogger, output: TextIOWrapper, expected: str):
        log.setLevel("info")
        log.info("Bär: 🐻")
        assert getvalue(output) == f"[test][info] {expected}\n"

    def test_logstream_switch(self, log: StreamlinkLogger, output: TextIOWrapper):
        output_ascii = TextIOWrapper(BytesIO(), encoding="ascii", errors="strict")
        log.setLevel("info")

        log.info("Bär: 🐻")
        assert getvalue(output) == "[test][info] Bär: 🐻\n"
        assert getvalue(output_ascii) == ""

        handler = log.handlers[0]
        assert isinstance(handler, logging.StreamHandler)
        handler.setStream(output_ascii)  # type: ignore  # Expected: Never ???

        log.info("Bär: 🐻")
        assert getvalue(output) == "[test][info] Bär: 🐻\n"
        assert getvalue(output_ascii) == "[test][info] B\\xe4r: \\U0001f43b\n"

    @pytest.mark.parametrize("log", [pytest.param({"stream": None}, id="no-stdout")], indirect=["log"])
    def test_no_stream(self, log: StreamlinkLogger):
        assert not log.handlers

    @pytest.mark.parametrize(
        "errno",
        [
            pytest.param(EPIPE, id="EPIPE", marks=pytest.mark.posix_only),
            pytest.param(EINVAL, id="EINVAL", marks=pytest.mark.windows_only),
        ],
    )
    def test_brokenpipeerror(
        self,
        monkeypatch: pytest.MonkeyPatch,
        capsys: pytest.CaptureFixture[str],
        log: StreamlinkLogger,
        errno: int,
    ):
        def flush(*_, **__):
            exception = OSError()
            exception.errno = errno
            raise exception

        streamhandler = log.handlers[0]
        assert isinstance(streamhandler, logging.StreamHandler)
        monkeypatch.setattr(streamhandler.stream, "flush", flush)

        log.setLevel("info")
        log.info("foo")

        # logging.StreamHandler will write emit()/flush() errors to stderr via handleError()
        out, err = capsys.readouterr()
        assert not out
        assert not err

    def test_logfile(self, logfile: Path, log: StreamlinkLogger, output: TextIOWrapper):
        log.setLevel("info")
        log.info("Hello world, Γειά σου Κόσμε, こんにちは世界")  # noqa: RUF001
        log.handlers[0].flush()
        with logfile.open("r", encoding="utf-8") as fh:
            assert fh.read() == "[test][info] Hello world, Γειά σου Κόσμε, こんにちは世界\n"  # noqa: RUF001


class TestCaptureWarnings:
    @staticmethod
    def _warn(messages: Iterable[tuple[str, type[Warning]]], filterwarnings=None):
        frame = currentframe()
        assert frame
        assert frame.f_back
        lineno = getframeinfo(frame.f_back).lineno

        with warnings.catch_warnings():
            warnings.filterwarnings(filterwarnings or "always")
            for message, warningclass in messages:
                warnings.warn(message, warningclass, stacklevel=2)

        return lineno

    def test_no_capture(self, log: StreamlinkLogger, output: TextIOWrapper):
        with pytest.warns(UserWarning, match=r"^Test warning$"):
            self._warn([("Test warning", UserWarning)])
        assert getvalue(output) == ""

    @pytest.mark.parametrize("log", [{"capture_warnings": True}], indirect=["log"])
    @pytest.mark.parametrize(
        ("warning", "expected", "origin"),
        [
            (("Test warning", UserWarning), "[warnings][userwarning] Test warning\n", True),
            (("Test warning", DeprecationWarning), "[warnings][deprecationwarning] Test warning\n", True),
            (("Test warning", FutureWarning), "[warnings][futurewarning] Test warning\n", True),
            (("Test warning", StreamlinkWarning), "[warnings][streamlinkwarning] Test warning\n", False),
            (("Test warning", StreamlinkDeprecationWarning), "[warnings][streamlinkdeprecation] Test warning\n", False),
        ],
    )
    def test_capture(
        self,
        recwarn: pytest.WarningsRecorder,
        log: StreamlinkLogger,
        output: TextIOWrapper,
        warning: tuple[str, type[Warning]],
        expected: str,
        origin: bool,
    ):
        lineno = self._warn([warning])
        expected += f"  {__file__}:{lineno}\n" if origin else ""
        assert recwarn.list == []
        assert getvalue(output) == expected

    @pytest.mark.parametrize("log", [{"capture_warnings": True}], indirect=["log"])
    def test_capture_logrecord(
        self,
        recwarn: pytest.WarningsRecorder,
        caplog: pytest.LogCaptureFixture,
        log: StreamlinkLogger,
    ):
        lineno = self._warn([("Test warning", UserWarning)])
        path = Path(__file__)
        assert recwarn.list == []
        assert [
            (
                r.name,
                r.levelname,
                r.pathname,
                r.filename,
                r.module,
                r.lineno,
                r.message,
            )
            for r in caplog.records
        ] == [
            (
                "warnings",
                "userwarning",
                __file__,
                path.name,
                path.stem,
                lineno,
                f"Test warning\n  {__file__}:{lineno}",
            ),
        ]

    @pytest.mark.parametrize("log", [{"capture_warnings": True}], indirect=["log"])
    def test_capture_consecutive(
        self,
        recwarn: pytest.WarningsRecorder,
        log: StreamlinkLogger,
        output: TextIOWrapper,
    ):
        lineno = self._warn([("foo", DeprecationWarning), ("bar", FutureWarning)])
        assert recwarn.list == []
        assert getvalue(output) == (
            f"[warnings][deprecationwarning] foo\n  {__file__}:{lineno}\n"
            + f"[warnings][futurewarning] bar\n  {__file__}:{lineno}\n"
        )

    @pytest.mark.parametrize("log", [{"capture_warnings": True}], indirect=["log"])
    def test_capture_consecutive_once(
        self,
        recwarn: pytest.WarningsRecorder,
        log: StreamlinkLogger,
        output: TextIOWrapper,
    ):
        lineno = self._warn([("foo", UserWarning), ("foo", UserWarning)], "once")
        assert recwarn.list == []
        assert getvalue(output) == f"[warnings][userwarning] foo\n  {__file__}:{lineno}\n"

    @pytest.mark.parametrize("log", [{"capture_warnings": True}], indirect=["log"])
    @pytest.mark.parametrize(
        "warning",
        [
            ("Test warning", UserWarning),
            ("Test warning", DeprecationWarning),
            ("Test warning", FutureWarning),
        ],
    )
    def test_ignored(
        self,
        recwarn: pytest.WarningsRecorder,
        log: StreamlinkLogger,
        output: TextIOWrapper,
        warning: tuple[str, type[Warning]],
    ):
        self._warn([warning], filterwarnings="ignore")
        assert recwarn.list == []
        assert getvalue(output) == ""



from streamlink.plugins.pluzz import Pluzz
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlPluzz(PluginCanHandleUrl):
    __plugin__ = Pluzz

    should_match = [
        ("francetv", "https://www.france.tv/france-2/direct.html"),
        ("francetv", "https://www.france.tv/france-3/direct.html"),
        ("francetv", "https://www.france.tv/france-4/direct.html"),
        ("francetv", "https://www.france.tv/france-5/direct.html"),
        ("francetv", "https://www.france.tv/franceinfo/direct.html"),
        ("francetv", "https://www.france.tv/france-2/journal-20h00/141003-edition-du-lundi-8-mai-2017.html"),
        (
            "francetvinfofr",
            "https://www.francetvinfo.fr/en-direct/tv.html",
        ),
        (
            "francetvinfofr",
            "https://www.francetvinfo.fr/meteo/orages/inondations-dans-le-gard-plus-de-deux-mois-de-pluie-en-quelques-heures-des-degats-mais-pas-de-victime_4771265.html",
        ),
        (
            "francetvinfofr",
            "https://france3-regions.francetvinfo.fr/bourgogne-franche-comte/direct/franche-comte",
        ),
        (
            "francetvinfofr",
            "https://la1ere.francetvinfo.fr/programme-video/france-3_outremer-ledoc/diffusion/2958951-polynesie-les-sages-de-l-ocean.html",
        ),
        (
            "francetvinfofr",
            "https://la1ere.francetvinfo.fr/info-en-continu-24-24",
        ),
    ]



from streamlink.plugins.welt import Welt
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlWelt(PluginCanHandleUrl):
    __plugin__ = Welt

    should_match = [
        "https://welt.de/tv-programm-live-stream/",
        "https://www.welt.de/tv-programm-live-stream/",
        "https://www.welt.de/tv-programm-n24-doku/",
        "https://www.welt.de/mediathek/dokumentation/space/strip-the-cosmos/sendung192055593/Strip-the-Cosmos-Die-Megastuerme-der-Planeten.html",
        "https://www.welt.de/mediathek/magazin/gesellschaft/sendung247758214/Die-Welt-am-Wochenende-Auf-hoher-See-mit-der-Gorch-Fock.html",
    ]



from __future__ import annotations

import re
from textwrap import dedent
from typing import TYPE_CHECKING

import pytest

import streamlink_cli.main

# noinspection PyProtectedMember
from streamlink.plugin.plugin import HIGH_PRIORITY, NORMAL_PRIORITY, Matcher, Matchers, Plugin, pluginmatcher


if TYPE_CHECKING:
    from streamlink.session import Streamlink


@pytest.fixture(autouse=True)
def session(session: Streamlink):
    @pluginmatcher(re.compile(r"foo", re.IGNORECASE))
    @pluginmatcher(name="asdf", pattern=re.compile(r"bar"), priority=HIGH_PRIORITY)
    @pluginmatcher(re.compile(rf"baz{chr(0x0A)}qux", re.IGNORECASE | re.VERBOSE), priority=100)
    class MyPlugin(Plugin):
        def _get_streams(self):  # pragma: no cover
            pass

    session.plugins.update({"plugin": MyPlugin})

    # noinspection PyProtectedMember
    session.plugins._matchers["lazy"] = Matchers(
        Matcher(re.compile(r"FOO"), priority=NORMAL_PRIORITY),
    )

    return session


@pytest.mark.parametrize(
    ("argv", "output"),
    [
        pytest.param(
            ["--show-matchers", "doesnotexist"],
            "error: Plugin not found\n",
            id="no-json",
        ),
        pytest.param(
            ["--show-matchers", "doesnotexist", "--json"],
            """{\n  "error": "Plugin not found"\n}\n""",
            id="json",
        ),
    ],
    indirect=["argv"],
)
def test_plugin_not_found(capsys: pytest.CaptureFixture[str], argv: list[str], output: str):
    with pytest.raises(SystemExit) as exc_info:
        streamlink_cli.main.main()
    assert exc_info.value.code == 1
    stdout, stderr = capsys.readouterr()
    assert stdout == output
    assert stderr == ""


@pytest.mark.parametrize(
    ("argv", "output"),
    [
        pytest.param(
            ["--show-matchers", "plugin"],
            dedent("""
                - flags: IGNORECASE
                  pattern: foo
                - name: asdf
                  priority: HIGH
                  pattern: bar
                - priority: 100
                  flags: IGNORECASE & VERBOSE
                  pattern:
                    baz
                    qux
            """).lstrip(),
            id="non-lazy-no-json",
        ),
        pytest.param(
            ["--show-matchers", "lazy"],
            dedent("""
                - pattern: FOO
            """).lstrip(),
            id="lazy-no-json",
        ),
        pytest.param(
            ["--show-matchers", "plugin", "--json"],
            dedent(f"""
                [
                  {{
                    "name": null,
                    "priority": {NORMAL_PRIORITY},
                    "flags": {int(re.IGNORECASE)},
                    "pattern": "foo"
                  }},
                  {{
                    "name": "asdf",
                    "priority": {HIGH_PRIORITY},
                    "flags": 0,
                    "pattern": "bar"
                  }},
                  {{
                    "name": null,
                    "priority": 100,
                    "flags": {int(re.IGNORECASE + re.VERBOSE)},
                    "pattern": "baz\\nqux"
                  }}
                ]
            """).lstrip(),
            id="json",
        ),
        pytest.param(
            ["--show-matchers", "lazy", "--json"],
            dedent(f"""
                [
                  {{
                    "name": null,
                    "priority": {NORMAL_PRIORITY},
                    "flags": 0,
                    "pattern": "FOO"
                  }}
                ]
            """).lstrip(),
            id="non-lazy-json",
        ),
    ],
    indirect=["argv"],
)
def test_show_matchers(capsys: pytest.CaptureFixture[str], argv: list[str], output: str):
    with pytest.raises(SystemExit) as exc_info:
        streamlink_cli.main.main()
    assert exc_info.value.code == 0
    stdout, stderr = capsys.readouterr()
    assert stdout == output
    assert stderr == ""



from __future__ import annotations

import errno
from collections import deque
from pathlib import Path
from threading import Thread
from typing import TYPE_CHECKING
from unittest.mock import Mock, patch

import pytest

from streamlink.stream.stream import StreamIO
from streamlink_cli.console.progress import Progress
from streamlink_cli.output import FileOutput, HTTPOutput, PlayerOutput
from streamlink_cli.streamrunner import PlayerPollThread, StreamRunner, log as streamrunnerlogger
from tests.testutils.handshake import Handshake


if TYPE_CHECKING:
    from collections.abc import Callable


TIMEOUT_AWAIT_HANDSHAKE = 1
TIMEOUT_AWAIT_THREADJOIN = 1


class _TestableWithHandshake:
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.handshake = Handshake()


class EventedPlayerPollThread(_TestableWithHandshake, PlayerPollThread):
    POLLING_INTERVAL = 0

    def poll(self):
        with self.handshake():
            return super().poll()

    def close(self):
        super().close()
        # Let thread terminate on close()
        self.handshake.go()


class FakeStream(_TestableWithHandshake, StreamIO):
    """Fake stream implementation, for feeding sample data to the stream runner and simulating read pauses and read errors"""

    def __init__(self) -> None:
        super().__init__()
        self.data: deque[bytes | Callable[[], bytes]] = deque()

    # noinspection PyUnusedLocal
    def read(self, *args):
        with self.handshake():
            if not self.data:
                return b""
            data = self.data.popleft()
            if callable(data) and not isinstance(data, bytes):
                return data()
            else:
                return data


class FakeOutput(_TestableWithHandshake):
    """Common output/http-server/progress interface, for caching all write() calls and simulating write errors"""

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.data: list[bytes] = []

    def write(self, data):
        with self.handshake():
            return self._write(data)

    def _write(self, data):
        self.data.append(data)


class FakePlayerOutput(FakeOutput, PlayerOutput):
    def open(self):
        with patch("streamlink_cli.output.player.which", side_effect=lambda path: path):
            return super().open()


class FakeFileOutput(FakeOutput, FileOutput):
    pass


class FakeHTTPOutput(FakeOutput, HTTPOutput):
    pass


class FakeProgress(FakeOutput, Progress):
    # we're not interested in any application logic of the Progress class
    update = print_end = lambda *_, **__: None


class FakeStreamRunner(StreamRunner):
    # override and remove optional typing annotations
    playerpoller: EventedPlayerPollThread
    progress: FakeProgress


@pytest.fixture(autouse=True)
def _logging(caplog: pytest.LogCaptureFixture):
    assert streamrunnerlogger.name == "streamlink.cli"
    caplog.set_level(1, "streamlink")


@pytest.fixture()
def stream():
    stream = FakeStream()
    yield stream
    assert stream.closed


# "stream_runner" fixture dependency declared in downstream scopes
@pytest.fixture()
def runnerthread(request: pytest.FixtureRequest, stream_runner: StreamRunner):
    class RunnerThread(Thread):
        exception = None

        def run(self):
            try:
                super().run()
            except BaseException as err:
                self.exception = err

    thread = RunnerThread(
        daemon=True,
        name="Runner thread",
        target=stream_runner.run,
        args=(b"prebuffer",),
    )
    yield thread

    assert_thread_termination(thread, "Runner thread has terminated")

    exception = getattr(request, "param", {}).get("exception", None)
    assert isinstance(thread.exception, type(exception))
    assert str(thread.exception) == str(exception)


async def assert_handshake_steps(*items: _TestableWithHandshake) -> None:
    for item in items:
        assert await item.handshake.asyncstep(TIMEOUT_AWAIT_HANDSHAKE) is True


def assert_thread_termination(thread: Thread, assertion: str):
    thread.join(TIMEOUT_AWAIT_THREADJOIN)
    assert not thread.is_alive(), assertion


class TestPlayerOutput:
    @pytest.fixture()
    def player_process(self):
        player_process = Mock()
        player_process.poll = Mock(return_value=None)

        return player_process

    @pytest.fixture()
    def output(self, monkeypatch: pytest.MonkeyPatch, player_process: Mock):
        mock_popen = Mock(return_value=player_process)
        monkeypatch.setattr("subprocess.Popen", mock_popen)
        monkeypatch.setattr("streamlink_cli.output.player.sleep", Mock())

        output = FakePlayerOutput(Path("mocked"))
        output.open()
        try:
            yield output
        finally:
            output.close()

    @pytest.fixture()
    def stream_runner(self, monkeypatch: pytest.MonkeyPatch, stream: FakeStream, output: FakePlayerOutput):
        monkeypatch.setattr("streamlink_cli.streamrunner.PlayerPollThread", EventedPlayerPollThread)
        stream_runner = StreamRunner(stream, output)
        assert isinstance(stream_runner.playerpoller, EventedPlayerPollThread)
        assert not stream_runner.playerpoller.is_alive()
        assert not isinstance(stream_runner.output, HTTPOutput)
        assert not stream_runner.progress
        yield stream_runner
        assert not stream_runner.playerpoller.is_alive()

    @pytest.mark.trio()
    async def test_read_write(
        self,
        caplog: pytest.LogCaptureFixture,
        runnerthread: Thread,
        stream_runner: FakeStreamRunner,
        stream: FakeStream,
        output: FakePlayerOutput,
    ):
        stream.data.extend((b"foo", b"bar"))

        runnerthread.start()
        assert output.data == []

        # write prebuffer
        await assert_handshake_steps(output)
        assert output.data == [b"prebuffer"]

        # read and write next chunk
        await assert_handshake_steps(stream, output)
        assert output.data == [b"prebuffer", b"foo"]

        # poll player process
        await assert_handshake_steps(stream_runner.playerpoller)
        assert stream_runner.playerpoller.is_alive()

        # read and write next chunk
        await assert_handshake_steps(stream, output)
        assert output.data == [b"prebuffer", b"foo", b"bar"]

        assert not stream.closed, "Stream is not closed"

        # read stream end
        await assert_handshake_steps(stream)
        assert output.data == [b"prebuffer", b"foo", b"bar"]

        # wait for runner thread to terminate first before asserting log records
        assert_thread_termination(runnerthread, "Runner thread has terminated")
        assert [(record.module, record.threadName, record.levelname, record.message) for record in caplog.records] == [
            ("streamrunner", "Runner thread", "info", "Stream ended"),
        ]

    @pytest.mark.trio()
    async def test_paused(
        self,
        caplog: pytest.LogCaptureFixture,
        runnerthread: Thread,
        stream_runner: FakeStreamRunner,
        stream: FakeStream,
        output: FakePlayerOutput,
    ):
        delayed = Handshake()

        def item():
            with delayed():
                return b"delayed"

        stream.data.append(item)

        runnerthread.start()
        assert output.data == []

        # write prebuffer
        await assert_handshake_steps(output)
        assert output.data == [b"prebuffer"]
        assert not delayed.wait_ready(0), "Delayed chunk has not been read yet"

        # attempt reading delayed chunk
        stream.handshake.go()
        assert delayed.wait_ready(TIMEOUT_AWAIT_HANDSHAKE), "read() call of delayed chunk is paused"
        assert output.data == [b"prebuffer"]

        assert not stream.closed, "Stream is not closed"

        # poll player process
        await assert_handshake_steps(stream_runner.playerpoller)
        assert stream_runner.playerpoller.is_alive()

        # unpause delayed chunk
        delayed.go()
        assert stream.handshake.wait_done(TIMEOUT_AWAIT_HANDSHAKE), "Delayed chunk has successfully been read"
        await assert_handshake_steps(output)
        assert output.data == [b"prebuffer", b"delayed"]

        assert not stream.closed, "Stream is not closed"

        # read stream end
        await assert_handshake_steps(stream)
        assert output.data == [b"prebuffer", b"delayed"]

        # wait for runner thread to terminate first before asserting log records
        assert_thread_termination(runnerthread, "Runner thread has terminated")
        assert [(record.module, record.threadName, record.levelname, record.message) for record in caplog.records] == [
            ("streamrunner", "Runner thread", "info", "Stream ended"),
        ]

    @pytest.mark.trio()
    @pytest.mark.parametrize(
        ("writeerror", "runnerthread"),
        [
            pytest.param(
                OSError(errno.EPIPE, "Broken pipe"),
                {},
                id="Acceptable error: EPIPE",
            ),
            pytest.param(
                OSError(errno.EINVAL, "Invalid argument"),
                {},
                id="Acceptable error: EINVAL",
            ),
            pytest.param(
                OSError(errno.ECONNRESET, "Connection reset"),
                {},
                id="Acceptable error: ECONNRESET",
            ),
            pytest.param(
                OSError("Unknown error"),
                {"exception": OSError("Error when writing to output: Unknown error, exiting")},
                id="Non-acceptable error",
            ),
        ],
        indirect=["runnerthread"],
    )
    async def test_player_close(
        self,
        caplog: pytest.LogCaptureFixture,
        runnerthread: Thread,
        stream_runner: FakeStreamRunner,
        stream: FakeStream,
        output: FakePlayerOutput,
        player_process: Mock,
        writeerror: Exception,
    ):
        stream.data.extend((b"foo", b"bar"))

        runnerthread.start()
        assert output.data == []

        # write prebuffer
        await assert_handshake_steps(output)
        assert output.data == [b"prebuffer"]

        # poll player process
        await assert_handshake_steps(stream_runner.playerpoller)
        assert stream_runner.playerpoller.is_alive()

        # read and write next chunk
        await assert_handshake_steps(stream, output)
        assert output.data == [b"prebuffer", b"foo"]

        assert not stream.closed, "Stream is not closed yet"

        # close player
        with patch.object(output, "_write", side_effect=writeerror):
            # let player process terminate with code 0 and poll process once
            player_process.poll.return_value = 0
            await assert_handshake_steps(stream_runner.playerpoller)
            assert_thread_termination(stream_runner.playerpoller, "Polling has stopped after player process terminated")

            assert stream.closed, "Stream got closed after the player was closed"

            # read and write next chunk (write will now also raise)
            await assert_handshake_steps(stream, output)
            assert output.data == [b"prebuffer", b"foo"]

        # wait for runner thread to terminate first before asserting log records
        assert_thread_termination(runnerthread, "Runner thread has terminated")
        assert [(record.module, record.threadName, record.levelname, record.message) for record in caplog.records] == [
            ("streamrunner", "EventedPlayerPollThread", "info", "Player closed"),
            ("streamrunner", "Runner thread", "info", "Stream ended"),
        ]

    @pytest.mark.trio()
    async def test_player_close_paused(
        self,
        caplog: pytest.LogCaptureFixture,
        runnerthread: Thread,
        stream_runner: FakeStreamRunner,
        stream: FakeStream,
        output: FakePlayerOutput,
        player_process: Mock,
    ):
        delayed = Handshake()

        def item():
            with delayed():
                return b""

        stream.data.append(item)

        runnerthread.start()
        assert output.data == []

        # write prebuffer
        await assert_handshake_steps(output)
        assert output.data == [b"prebuffer"]
        assert not delayed.wait_ready(0), "Delayed chunk has not been read yet"

        # poll player process
        await assert_handshake_steps(stream_runner.playerpoller)
        assert stream_runner.playerpoller.is_alive()

        stream.handshake.go()
        assert delayed.wait_ready(TIMEOUT_AWAIT_HANDSHAKE), "read() call of delayed chunk is paused"
        assert output.data == [b"prebuffer"]

        assert not stream.closed, "Stream is not closed yet"

        # let player process terminate with code 0 and poll process once
        player_process.poll.return_value = 0
        await assert_handshake_steps(stream_runner.playerpoller)
        assert_thread_termination(stream_runner.playerpoller, "Polling has stopped after player process terminated")

        assert stream.closed, "Stream got closed after the player was closed, even if the stream was paused"

        # unpause delayed chunk (stream end)
        delayed.go()
        assert stream.handshake.wait_done(TIMEOUT_AWAIT_HANDSHAKE), "Delayed chunk has successfully been read"
        assert output.data == [b"prebuffer"]

        # wait for runner thread to terminate first before asserting log records
        assert_thread_termination(runnerthread, "Runner thread has terminated")
        assert [(record.module, record.threadName, record.levelname, record.message) for record in caplog.records] == [
            ("streamrunner", "EventedPlayerPollThread", "info", "Player closed"),
            ("streamrunner", "Runner thread", "info", "Stream ended"),
        ]

    @pytest.mark.trio()
    @pytest.mark.parametrize(
        "runnerthread",
        [{"exception": OSError("Error when reading from stream: Read timeout, exiting")}],
        indirect=["runnerthread"],
    )
    async def test_readerror(
        self,
        caplog: pytest.LogCaptureFixture,
        runnerthread: Thread,
        stream_runner: FakeStreamRunner,
        stream: FakeStream,
        output: FakePlayerOutput,
    ):
        # make next read() call raise a read-timeout error
        stream.data.append(Mock(side_effect=OSError("Read timeout")))

        runnerthread.start()
        assert output.data == []

        # write prebuffer
        await assert_handshake_steps(output)
        assert output.data == [b"prebuffer"]

        # poll player process
        await assert_handshake_steps(stream_runner.playerpoller)
        assert stream_runner.playerpoller.is_alive()

        # read stream (will raise a read timeout)
        await assert_handshake_steps(stream)

        # poll player process again
        await assert_handshake_steps(stream_runner.playerpoller)
        assert_thread_termination(stream_runner.playerpoller, "Polling has stopped on read error")

        # wait for runner thread to terminate first before asserting log records
        assert_thread_termination(runnerthread, "Runner thread has terminated")
        assert [(record.module, record.threadName, record.levelname, record.message) for record in caplog.records] == [
            ("streamrunner", "Runner thread", "info", "Stream ended"),
        ]


class TestHTTPServer:
    @pytest.fixture()
    def output(self):
        return FakeHTTPOutput()

    @pytest.fixture()
    def stream_runner(self, stream: FakeStream, output: FakeHTTPOutput):
        stream_runner = StreamRunner(stream, output)
        assert not stream_runner.playerpoller
        assert not stream_runner.progress
        assert isinstance(stream_runner.output, HTTPOutput)
        return stream_runner

    @pytest.mark.trio()
    async def test_read_write(
        self,
        caplog: pytest.LogCaptureFixture,
        runnerthread: Thread,
        stream_runner: FakeStreamRunner,
        stream: FakeStream,
        output: FakeHTTPOutput,
    ):
        stream.data.extend((b"foo", b"bar"))

        runnerthread.start()
        assert output.data == []

        # write prebuffer
        await assert_handshake_steps(output)
        assert output.data == [b"prebuffer"]

        # read and write next chunk
        await assert_handshake_steps(stream, output)
        assert output.data == [b"prebuffer", b"foo"]

        # read and write next chunk
        await assert_handshake_steps(stream, output)
        assert output.data == [b"prebuffer", b"foo", b"bar"]

        assert not stream.closed, "Stream is not closed"

        # read stream end
        await assert_handshake_steps(stream)
        assert output.data == [b"prebuffer", b"foo", b"bar"]

        # wait for runner thread to terminate first before asserting log records
        assert_thread_termination(runnerthread, "Runner thread has terminated")
        assert [(record.module, record.threadName, record.levelname, record.message) for record in caplog.records] == [
            ("streamrunner", "Runner thread", "info", "Stream ended"),
        ]

    @pytest.mark.parametrize(
        ("writeerror", "logs", "runnerthread"),
        [
            pytest.param(
                OSError(errno.EPIPE, "Broken pipe"),
                True,
                {},
                id="Acceptable error: EPIPE",
            ),
            pytest.param(
                OSError(errno.EINVAL, "Invalid argument"),
                True,
                {},
                id="Acceptable error: EINVAL",
            ),
            pytest.param(
                OSError(errno.ECONNRESET, "Connection reset"),
                True,
                {},
                id="Acceptable error: ECONNRESET",
            ),
            pytest.param(
                OSError("Unknown error"),
                False,
                {"exception": OSError("Error when writing to output: Unknown error, exiting")},
                id="Non-acceptable error",
            ),
        ],
        indirect=["runnerthread"],
    )
    def test_writeerror(
        self,
        caplog: pytest.LogCaptureFixture,
        runnerthread: Thread,
        stream_runner: FakeStreamRunner,
        stream: FakeStream,
        output: FakePlayerOutput,
        logs: bool,
        writeerror: Exception,
    ):
        runnerthread.start()

        with patch.object(output, "_write", side_effect=writeerror):
            assert output.handshake.step(TIMEOUT_AWAIT_HANDSHAKE)
            assert output.data == []

        # wait for runner thread to terminate first before asserting log records
        assert_thread_termination(runnerthread, "Runner thread has terminated")
        assert [(record.module, record.threadName, record.levelname, record.message) for record in caplog.records] == [
            *([("streamrunner", "Runner thread", "info", "HTTP connection closed")] if logs else []),
            ("streamrunner", "Runner thread", "info", "Stream ended"),
        ]


class TestProgress:
    @pytest.fixture()
    def output(self):
        return FakeFileOutput(Path("filename"))

    @pytest.fixture()
    def progress(self):
        return FakeProgress(console=Mock(), path=Path("filename"))

    @pytest.fixture()
    def stream_runner(self, stream: FakeStream, output: FakeFileOutput, progress: FakeProgress):
        stream_runner = FakeStreamRunner(stream, output, progress=progress)
        assert not stream_runner.playerpoller
        assert stream_runner.progress is progress
        assert not stream_runner.progress.is_alive()
        yield stream_runner
        assert not stream_runner.progress.is_alive()

    @pytest.mark.trio()
    async def test_read_write(
        self,
        caplog: pytest.LogCaptureFixture,
        runnerthread: Thread,
        stream_runner: FakeStreamRunner,
        stream: FakeStream,
        output: FakeFileOutput,
    ):
        stream.data.extend((b"foo", b"bar"))

        runnerthread.start()
        assert output.data == []

        # write prebuffer
        await assert_handshake_steps(output, stream_runner.progress)
        assert output.data == [b"prebuffer"]
        assert stream_runner.progress.data == [b"prebuffer"]

        # read and write next chunk
        await assert_handshake_steps(stream, output, stream_runner.progress)
        assert output.data == [b"prebuffer", b"foo"]
        assert stream_runner.progress.data == [b"prebuffer", b"foo"]

        # read and write next chunk
        await assert_handshake_steps(stream, output, stream_runner.progress)
        assert output.data == [b"prebuffer", b"foo", b"bar"]
        assert stream_runner.progress.data == [b"prebuffer", b"foo", b"bar"]

        assert not stream.closed, "Stream is not closed"

        # read stream end
        await assert_handshake_steps(stream)
        assert output.data == [b"prebuffer", b"foo", b"bar"]
        assert stream_runner.progress.data == [b"prebuffer", b"foo", b"bar"]

        # wait for runner thread to terminate first before asserting log records
        assert_thread_termination(runnerthread, "Runner thread has terminated")
        assert [(record.module, record.threadName, record.levelname, record.message) for record in caplog.records] == [
            ("streamrunner", "Runner thread", "info", "Stream ended"),
        ]



import argparse
from datetime import datetime, timedelta, timezone

import freezegun
import isodate  # type: ignore[import]
import pytest

from streamlink.utils.times import (
    LOCAL,
    UTC,
    fromlocaltimestamp,
    fromtimestamp,
    hours_minutes_seconds,
    hours_minutes_seconds_float,
    localnow,
    now,
    parse_datetime,
    seconds_to_hhmmss,
)


class TestDatetime:
    TS_Y2K = 946684800

    @pytest.fixture()
    def chatham_islands(self, monkeypatch: pytest.MonkeyPatch):
        chatham_islands = timezone(timedelta(hours=12, minutes=45))
        monkeypatch.setattr("streamlink.utils.times.LOCAL", chatham_islands)
        return chatham_islands

    def test_constants(self):
        assert UTC is timezone.utc
        assert LOCAL is isodate.LOCAL

    def test_now(self):
        with freezegun.freeze_time("2000-01-01T00:00:00Z"):
            assert now() == datetime(2000, 1, 1, 0, 0, 0, 0, timezone.utc)

    def test_localnow(self, chatham_islands: timezone):
        with freezegun.freeze_time("2000-01-01T00:00:00+1245"):
            assert localnow() == datetime(2000, 1, 1, 0, 0, 0, 0, chatham_islands)

    def test_fromtimestamp(self):
        assert fromtimestamp(self.TS_Y2K) == datetime(2000, 1, 1, 0, 0, 0, 0, timezone.utc)

    def test_fromlocaltimestamp(self, chatham_islands: timezone):
        assert fromlocaltimestamp(self.TS_Y2K) == datetime(2000, 1, 1, 12, 45, 0, 0, chatham_islands)
        assert fromlocaltimestamp(self.TS_Y2K) == datetime(2000, 1, 1, 0, 0, 0, 0, timezone.utc)

    def test_parse_datetime(self, chatham_islands: timezone):
        assert parse_datetime("2000-01-01T00:00:00") == datetime(2000, 1, 1, 0, 0, 0, 0)  # noqa: DTZ001
        assert parse_datetime("2000-01-01T00:00:00Z") == datetime(2000, 1, 1, 0, 0, 0, 0, timezone.utc)
        assert parse_datetime("2000-01-01T00:00:00+1245") == datetime(2000, 1, 1, 0, 0, 0, 0, chatham_islands)
        with pytest.raises(isodate.ISO8601Error):
            parse_datetime("2000-01-01")


class TestHoursMinutesSeconds:
    @pytest.mark.parametrize(
        ("sign", "factor"),
        [
            pytest.param("", 1, id="positive"),
            pytest.param("-", -1, id="negative"),
        ],
    )
    @pytest.mark.parametrize(
        ("timestamp", "as_float", "expected"),
        [
            # decimals
            pytest.param("0", True, 0.0, id="zero"),
            pytest.param("123", True, 123.0, id="decimal without fraction"),
            pytest.param("123.456789", True, 123.456789, id="decimal with fraction"),
            # XX:XX:XX
            pytest.param("0:0", True, 0.0, id="0:0"),
            pytest.param("0:0:0", True, 0.0, id="0:0:0"),
            pytest.param("1:2", True, 62.0, id="X:X"),
            pytest.param("1:2.3", True, 62.3, id="X:X.X"),
            pytest.param("12:34", True, 754.0, id="XX:XX"),
            pytest.param("12:34.56", True, 754.56, id="XX:XX.XX"),
            pytest.param("123:45", True, 7425.0, id="XXX:XX"),
            pytest.param("123:45.67", True, 7425.67, id="XXX:XX.XX"),
            pytest.param("1:2:3", True, 3723.0, id="X:X:X"),
            pytest.param("1:2:3.4", True, 3723.4, id="X:X:X.X"),
            pytest.param("12:34:56", True, 45296.0, id="XX:XX:XX"),
            pytest.param("12:34:56.78", True, 45296.78, id="XX:XX:XX.XX"),
            pytest.param("123:4:5", True, 443045.0, id="XXX:X:X"),
            pytest.param("123:4:5.6", True, 443045.6, id="XXX:X:X.X"),
            # XXhXXmXXs
            pytest.param("0s", True, 0.0, id="0s"),
            pytest.param("0m0s", True, 0.0, id="0m0s"),
            pytest.param("0h0m0s", True, 0.0, id="0h0m0s"),
            pytest.param("1s", True, 1.0, id="Xs"),
            pytest.param("1.2s", True, 1.2, id="X.Xs"),
            pytest.param("12s", True, 12.0, id="XXs"),
            pytest.param("12.3s", True, 12.3, id="XX.Xs"),
            pytest.param("123s", True, 123.0, id="XXXs"),
            pytest.param("123.4s", True, 123.4, id="XXX.Xs"),
            pytest.param("1m", True, 60.0, id="Xm"),
            pytest.param("12m", True, 720.0, id="XXm"),
            pytest.param("123m", True, 7380.0, id="XXXm"),
            pytest.param("1h", True, 3600.0, id="Xh"),
            pytest.param("12h", True, 43200.0, id="XXh"),
            pytest.param("123h", True, 442800.0, id="XXXh"),
            pytest.param("1m2s", True, 62.0, id="XmXs"),
            pytest.param("1m2.3s", True, 62.3, id="XmX.Xs"),
            pytest.param("12m3s", True, 723.0, id="XXmXs"),
            pytest.param("12m3.4s", True, 723.4, id="XXmX.Xs"),
            pytest.param("12m34s", True, 754.0, id="XXmXXs"),
            pytest.param("12m34.5s", True, 754.5, id="XXmXX.Xs"),
            pytest.param("123m45s", True, 7425.0, id="XXXmXXs"),
            pytest.param("123m45.6s", True, 7425.6, id="XXXmXX.Xs"),
            pytest.param("1h2m3s", True, 3723.0, id="XhXmXs"),
            pytest.param("12h34m56s", True, 45296.0, id="XXhXXmXXs"),
            pytest.param("12h34m56.78s", True, 45296.78, id="XXhXXmXX.XXs"),
            pytest.param("123h4m5s", True, 443045.0, id="XXXhXmXs"),
            pytest.param("123h4m5.6s", True, 443045.6, id="XXXhXmX.Xs"),
            pytest.param("1h2s", True, 3602.0, id="XhXs"),
            pytest.param("1h2m", True, 3720.0, id="XhXs"),
            pytest.param("12.34S", True, 12.34, id="XX.XXS"),
            pytest.param("12M34.56S", True, 754.56, id="XXMXX.XXS"),
            pytest.param("12H34M56.78S", True, 45296.78, id="XXHXXMXX.XXS"),
            # integers
            pytest.param("0", False, 0, id="zero (int)"),
            pytest.param("123", False, 123, id="decimal without fraction (int)"),
            pytest.param("123.456789", False, 123, id="decimal with fraction (int)"),
            pytest.param("12:34:56", False, 45296, id="XX:XX:XX (int)"),
            pytest.param("12:34:56.78", False, 45296, id="XX:XX:XX.XX (int)"),
            pytest.param("12h34m56s", False, 45296, id="XXhXXmXXs (int)"),
            pytest.param("12h34m56.78s", False, 45296, id="XXhXXmXX.XXs (int)"),
            # base 10
            pytest.param("0123", True, 123.0, id="base10"),
            pytest.param("08:08:08", True, 29288.0, id="XX:XX:XX base10"),
            pytest.param("08h08m08s", True, 29288.0, id="XXhXXmXXs base10"),
        ],
    )
    def test_hours_minutes_seconds(self, timestamp: str, as_float: bool, sign: str, factor: int, expected: float):
        method = hours_minutes_seconds_float if as_float else hours_minutes_seconds
        res = method(f"{sign}{timestamp}")
        assert type(res) is type(expected)
        assert res == factor * expected

    @pytest.mark.parametrize(
        "timestamp",
        [
            # missing timestamp
            pytest.param("", id="empty"),
            pytest.param(" ", id="whitespace"),
            # invalid numbers
            pytest.param("+123", id="plus sign"),
            pytest.param("1e10", id="exponent notation"),
            pytest.param("1_000", id="digit notation"),
            pytest.param("NaN", id="NaN"),
            pytest.param("infinity", id="infinity"),
            pytest.param("0xff", id="base16"),
            # invalid format
            pytest.param("foo", id="invalid input"),
            pytest.param(" 1:2:3 ", id="untrimmed input"),
            pytest.param(":1:2", id="missing hours value"),
            pytest.param("1::2", id="missing minutes value"),
            pytest.param("1:2:", id="missing seconds value"),
            pytest.param("foo:1:2", id="invalid hours"),
            pytest.param("1:foo:2", id="invalid minutes"),
            pytest.param("1:2:foo", id="invalid seconds"),
            pytest.param("1:60", id="seconds with two digits gte 60"),
            pytest.param("1:60:59", id="minutes with two digits gte 60"),
            pytest.param("1:234", id="minutes and seconds with three digits"),
            pytest.param("1:234:56", id="hours and minutes with three digits"),
            pytest.param("1:23:456", id="hours and seconds with three digits"),
            pytest.param("1h2", id="missing minutes or seconds suffix"),
            pytest.param("1m2", id="missing seconds suffix"),
            pytest.param("1.2h", id="hours fraction"),
            pytest.param("1.2m", id="minutes fraction"),
            pytest.param("1:2s", id="mixed format"),
            pytest.param("1h2:3", id="mixed format"),
            pytest.param("1:2:3s", id="mixed format"),
            pytest.param("1:2m3s", id="mixed format"),
        ],
    )
    def test_hours_minutes_seconds_exception(self, timestamp: str):
        with pytest.raises(ValueError):  # noqa: PT011
            hours_minutes_seconds_float(timestamp)

    @pytest.mark.parametrize("method", [hours_minutes_seconds, hours_minutes_seconds_float])
    def test_hours_minutes_seconds_argparse_failure(self, method):
        parser = argparse.ArgumentParser(exit_on_error=False)
        parser.add_argument("hms", type=method)
        with pytest.raises(argparse.ArgumentError) as exc_info:
            parser.parse_args(["invalid"])
        assert str(exc_info.value) == "argument hms: invalid hours_minutes_seconds value: 'invalid'", \
            "has the correct method name, so argparse errors are useful"  # fmt: skip

    def test_hours_minutes_seconds_hashable(self):
        assert hash(hours_minutes_seconds) != hash(hours_minutes_seconds_float)

    def test_seconds_to_hhmmss(self):
        assert seconds_to_hhmmss(0) == "00:00:00"
        assert seconds_to_hhmmss(1) == "00:00:01"
        assert seconds_to_hhmmss(60) == "00:01:00"
        assert seconds_to_hhmmss(3600) == "01:00:00"

        assert seconds_to_hhmmss(13997) == "03:53:17"
        assert seconds_to_hhmmss(13997.4) == "03:53:17.4"



from __future__ import annotations

from contextlib import nullcontext
from typing import TYPE_CHECKING

import pytest

from streamlink.exceptions import NoStreamsError
from streamlink.plugins.vk import VK
from tests.plugins import PluginCanHandleUrl


if TYPE_CHECKING:
    import requests_mock as rm

    from streamlink import Streamlink


does_not_raise = nullcontext()


class TestPluginCanHandleUrlVK(PluginCanHandleUrl):
    __plugin__ = VK

    should_match_groups = [
        # vod
        (("video", "https://vk.com/video-9944999_456239622"), {"id": "-9944999_456239622"}),
        (("video", "https://vk.ru/video-9944999_456239622"), {"id": "-9944999_456239622"}),
        # live
        (("video", "https://www.vk.com/video-143491903_456240010"), {"id": "-143491903_456240010"}),
        (("video", "https://www.vk.ru/video-143491903_456240010"), {"id": "-143491903_456240010"}),
        # video ID without leading dash
        (("video", "https://vk.com/video132886594_167211693"), {"id": "132886594_167211693"}),
        # video with required list param
        (("video", "https://vk.com/video-122792749_456242046?list=ln-YXS9u57K48LhkU6C3X"), {"id": "-122792749_456242046"}),
        # video/lives
        (("default", "https://vk.com/video/lives?z=video-143491903_456240010"), {}),
        # other paths / formats
        (("default", "https://vk.com/videos-24136539?z=video-24136539_456241155%2Fpl_-24136539_-2"), {}),
        (("default", "https://vk.com/video/@vesti?z=video-24136539_456241155%2Fpl_-24136539_-2"), {}),
        (("default", "https://vk.com/video?z=video-15755094_456245149%2Fpl_cat_lives"), {}),
        (("default", "https://vk.com/video?z=video132886594_167211693%2Fpl_cat_8"), {}),
        # search
        (("default", "https://vk.com/dota2?w=wall-126093223_1057924"), {}),
        (("default", "https://vk.com/search?c%5Bq%5D=dota&c%5Bsection%5D=auto&z=video295508334_171402661"), {}),
    ]

    should_not_match = [
        "https://vk.com/",
        "https://vk.ru/",
    ]


@pytest.mark.parametrize(
    ("url", "newurl", "raises"),
    [
        pytest.param(
            "https://vk.com/video-24136539_456241176",
            "https://vk.com/video-24136539_456241176",
            does_not_raise,
            id="canonical-video-url",
        ),
        pytest.param(
            "https://vk.com/videos-24136539?z=video-24136539_456241181%2Fpl_-24136539_-2",
            "https://vk.com/video-24136539_456241181",
            does_not_raise,
            id="z-querystring-parameter-with-pl",
        ),
        pytest.param(
            "https://vk.com/videos132886594?z=video132886594_167211693",
            "https://vk.com/video132886594_167211693",
            does_not_raise,
            id="z-querystring-parameter",
        ),
        pytest.param(
            "https://vk.com/search?c%5Bq%5D=dota&c%5Bsection%5D=auto&z=video295508334_171402661",
            "https://vk.com/video295508334_171402661",
            does_not_raise,
            id="search-z-querystring-parameter",
        ),
        pytest.param(
            "https://vk.com/dota2?w=wall-126093223_1057924",
            "https://vk.com/video-126093223_1057924",
            does_not_raise,
            id="from-meta-og-url",
        ),
        # no video ID
        pytest.param(
            "https://vk.com/videos-0123456789?z=",
            "",
            pytest.raises(NoStreamsError),
            id="no-streams-1",
        ),
        pytest.param(
            "https://vk.com/video/for_kids?z=video%2Fpl_-22277933_51053000",
            "",
            pytest.raises(NoStreamsError),
            id="no-streams-2",
        ),
    ],
)
def test_url_redirect(requests_mock: rm.Mocker, session: Streamlink, url: str, newurl: str, raises: nullcontext):
    # noinspection PyTypeChecker
    plugin: VK = VK(session, url)
    requests_mock.get(url, text=f"""<!DOCTYPE html><html><head><meta property="og:url" content="{newurl}"/></head></html>""")
    with raises:
        plugin.follow_vk_redirect()
        assert plugin.url == newurl
